"""
# Spectre-Format Netlister
"""

# Std-Lib Imports
from typing import Union

# Local Imports
import vlsir
import vlsir.circuit_pb2 as vckt
import vlsir.spice_pb2 as vsp

# Import the base-class
from .spectre_spice_shared import SpectreSpiceShared
from .base import ResolvedModule, ResolvedParams, SpiceModelRef, SpiceBuiltin, SpiceType


def map_primitive(rmodule: SpiceBuiltin, paramvals: ResolvedParams) -> str:
    """Map a primitive into Spectre's supported names and parameters.
    Returns the "apparent module name" for instances of the primitive.
    Argument `paramvals` is often modified along the way.

    Note spectre syntax is such that the "apparent module name" can be either of:
    (a) a fixed name per spice-prefix, for basic types (r, c, l, etc), or
    (b) the *model* name for model-based devices, (mos, bjt, diode, etc)"""

    # For voltage sources, add spectre's "type" parameter, and potentially rename several parameters
    if rmodule.spice_type == SpiceType.VSOURCE:
        vname: str = rmodule.module.name.name
        vtypes = dict(
            vdc="dc",
            vpulse="pulse",
            vsin="sine",
        )
        if vname not in vtypes:
            msg = f"Invalid or unsupported voltage-source type {vname}"
            raise ValueError(msg)
        paramvals.set("type", vtypes[vname])

        if vname == "vpulse":
            # For pulse sources, need to rename most parameters
            paramvals.rename("v1", "val0")
            paramvals.rename("v2", "val1")
            paramvals.rename("td", "delay")
            paramvals.rename("tr", "rise")
            paramvals.rename("tf", "fall")
            paramvals.rename("tpw", "width")
            paramvals.rename("tper", "period")
        elif vname == "vdc":
            if "ac" in paramvals:
                paramvals.rename("ac", "mag")

    # Mapping from spice-prefix to spectre-name for fixed-name types
    basics = {
        SpiceType.RESISTOR: "resistor",
        SpiceType.CAPACITOR: "capacitor",
        SpiceType.INDUCTOR: "inductor",
        SpiceType.VSOURCE: "vsource",
        SpiceType.ISOURCE: "isource",
        SpiceType.VCVS: "vcvs",
        SpiceType.VCCS: "vccs",
        SpiceType.CCCS: "cccs",
        SpiceType.CCVS: "ccvs",
    }
    if rmodule.spice_type in basics:
        return basics[rmodule.spice_type]

    # Other primitive-types get an "apparent module name" equal to their *model* name.
    model_based = {
        SpiceType.MOS,
        SpiceType.BIPOLAR,
        SpiceType.DIODE,
        SpiceType.TLINE,
    }
    if rmodule.spice_type in model_based:  # Invalid
        msg = f"Internal Error: Invalid model-based primitive {rmodule} should have been resolved to a model"
        raise RuntimeError(msg)

    # Otherwise, unclear what this is or how we got here.
    raise RuntimeError(f"Unsupported or Invalid Primitive {rmodule}")


class SpectreNetlister(SpectreSpiceShared):
    """Spectre-Format Netlister"""

    @property
    def enum(self):
        """Get our entry in the `NetlistFormat` enumeration"""
        from . import NetlistFormat

        return NetlistFormat.SPECTRE

    def write_module_definition(self, module: vckt.Module) -> None:
        """Create a Spectre-format definition for proto-Module `module`"""

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Collect and index vckt.Signals in this Module by name.
        self.collect_signals_by_name(module)

        # Create the sub-circuit definition header
        self.writeln(f"subckt {module_name} ")
        self.indent += 1

        if module.ports:  # Create its ports
            self.writeln(
                "+ "
                + " ".join([self.format_port_decl(pport) for pport in module.ports])
                + " "
            )
        else:
            self.writeln("+ // No ports ")

        # Create its parameters, if defined
        if module.parameters:
            formatted = [self.format_param_decl(pparam) for pparam in module.parameters]
            formatted = " ".join(formatted)
            self.writeln("parameters " + formatted + " ")
        else:
            self.writeln("+ // No parameters ")

        self.writeln("")  # End "header" facets
        # Note nothing need be done for internal signals;
        # spice and spectre create these "out of thin air"

        # Create its instances
        for pinst in module.instances:
            self.write_instance(pinst)
        self.writeln("")

        # Write any netlist-literal content
        self.write_literals(module.literals)

        # Close up the sub-circuit
        self.indent -= 1
        self.writeln("ends \n")

    def write_instance(self, pinst: vckt.Instance) -> None:
        """Create and return a netlist-string for Instance `pinst`"""

        # Initial resolution phase.
        # Start by getting the Module or ExternalModule definition.
        ref = self.resolve_reference(pinst.module)
        module = ref.module

        # Resolve its parameter values, including applying module-level defaults
        resolved_instance_parameters = self.get_instance_params(pinst, module)

        # And sort out its "apparent module name" for netlisting
        if isinstance(ref, ResolvedModule):
            module_name = ref.module_name
        elif isinstance(ref, SpiceModelRef):
            # Spectre format makes "model_name" and "module_name" look identical
            module_name = ref.model_name
        elif isinstance(ref, SpiceBuiltin):
            # Map the primitive into Spectre's built-in element names and parameters
            module_name = map_primitive(ref, resolved_instance_parameters)
        else:
            raise RuntimeError(f"Invalid reference {ref}")

        # OK now we have everything we need to *write* the instance.
        # Create the instance name
        # Note spectre format does not include any `SpiceType` based prefixing
        self.writeln(pinst.name + "")

        if module.ports:
            self.writeln("+ // Ports: ")
            # Get `module`'s port-order
            port_order = [pport.signal for pport in module.ports]
            # And write the Instance ports, in that order
            pconns = []
            connection_targets = {
                connection.portname: connection.target
                for connection in pinst.connections
            }
            for pname in port_order:
                pconn = connection_targets.get(pname, None)
                if pconn is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                pconns.append(pconn)
            self.writeln(
                "+ ( "
                + " ".join([self.format_connection_target(pconn) for pconn in pconns])
                + " )"
            )
        else:
            self.writeln("+ // No ports ")

        # Write the module-name
        self.writeln("+  " + module_name + " ")

        # Write the instance parameters
        self.write_instance_params(resolved_instance_parameters)

        # And add a post-instance blank line
        self.writeln("")

    def write_instance_params(self, pvals: ResolvedParams) -> None:
        """Write Instance parameters `pvals`"""
        if not pvals:
            return self.writeln("+ // No parameters ")
        # Write its parameter-values
        formatted = " ".join([f"{pname}={pval}" for pname, pval in pvals.items()])
        self.writeln("+  " + formatted + " ")

    def format_concat(self, pconc: vckt.Concat) -> str:
        """Format the Concatenation of several other Connections"""
        out = ""
        for part in pconc.parts:
            out += self.format_connection_target(part) + " "
        return out

    def format_port_decl(self, pport: vckt.Port) -> str:
        """Get a netlist `Port` definition"""
        # In Spectre, as well as most spice, this syntax is the same as referring to the Port.
        return self.format_port_ref(pport)

    def format_port_ref(self, pport: vckt.Port) -> str:
        """Get a netlist `Port` reference"""
        return self.format_signal_ref(self.get_signal(pport.signal))

    @classmethod
    def format_signal_ref(cls, psig: vckt.Signal) -> str:
        """Get a netlist definition for Signal `psig`"""
        if psig.width < 1:
            raise RuntimeError
        if psig.width == 1:  # width==1, i.e. a scalar signal
            return psig.name
        # Vector/ multi "bit" Signal. Creates several spice signals.
        return " ".join(
            [f"{psig.name}{cls.format_bus_bit(k)}" for k in reversed(range(psig.width))]
        )

    @classmethod
    def format_signal_slice(cls, pslice: vckt.Slice) -> str:
        """Get a netlist definition for Signal-Slice `pslice`"""
        base = pslice.signal
        indices = list(reversed(range(pslice.bot, pslice.top + 1)))
        if not len(indices):
            raise RuntimeError(f"Attempting to netlist empty slice {pslice}")
        return " ".join([f"{base}{cls.format_bus_bit(k)}" for k in indices])

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """Format-specific string-representation of a bus bit-index"""
        # Spectre netlisting uses an underscore prefix, e.g. `bus_0`
        return "_" + str(index)

    def write_comment(self, comment: str) -> None:
        """While Spectre *can* do a bunch of other comment-styles,
        the canonical one is generally the C-style line comment beginning with `//`."""
        self.writeln(f"// {comment}")

    @classmethod
    def format_sim_dut(cls, module_name: str) -> str:
        """# Format the top-level DUT instance for module name `module_name`."""
        return f"xtop 0 {module_name} // Top-Level DUT \n"

    def write_sim_header(self, inp: vsp.SimInput) -> None:
        """# Write header commentary for a `SimInput`"""

        # Note on this line here:
        self.writeln("simulator lang=spectre \n")
        # There's no need to write `simulator lang=spectre` in a file who's extension is `.scs`.
        # But I guess this doesn't hurt? And continues to work if the file gets copied to another extension?

        # Write the base-class version, it includes some nice commentary
        super().write_sim_header(inp)

        # FIXME: do we *really* want this global-zero?
        self.writeln("global 0")
        self.writeln("")

    def write_include(self, inc: vsp.Include) -> None:
        """# Write an `Include` statement"""
        self.writeln(f'include "{inc.path}"')

    def write_lib_include(self, lib: vsp.LibInclude) -> None:
        """# Write a `LibInclude` statement"""
        txt = f'include "{lib.path}" section={lib.section}'
        self.writeln(txt)

    def write_save(self, save: vsp.Save) -> None:
        """# Write a `Save` statement"""
        # FIXME!
        raise NotImplementedError(f"Unimplemented control card {save} for {self}")

    def write_meas(self, meas: vsp.Meas) -> None:
        """# Write a `Meas` statement"""
        # Measurements are written in Spice syntax; wrap them in "simulator lang".
        self.writeln(f"simulator lang=spice")
        txt = f".meas {meas.analysis_type} {meas.name} {meas.expr}"
        self.writeln(txt)
        self.writeln(f"simulator lang=spectre")

    def write_sim_param(self, param: vlsir.Param) -> None:
        """# Write a simulation-level parameter"""
        txt = f"parameters  {param.name}={self.get_param_value(param.value)}"
        self.writeln(txt)

    def write_sim_option(self, opt: vsp.SimOptions) -> None:
        """# Write a simulation option"""
        # FIXME: make this just `Param` instead
        raise NotImplementedError

    def write_ac(self, an: vsp.AcInput) -> None:
        """# Write an AC analysis."""

        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Unpack the analysis / sweep content
        fstart = an.fstart
        if fstart <= 0:
            raise ValueError(f"Invalid `fstart` {fstart}")
        fstop = an.fstop
        if fstop <= 0:
            raise ValueError(f"Invalid `fstop` {fstop}")
        npts = an.npts
        if npts <= 0:
            raise ValueError(f"Invalid `npts` {npts}")

        # Write the analysis command
        line = f"{an.analysis_name} ac start={fstart} stop={fstop} dec={npts}"
        self.writeln(line)

    def write_dc(self, an: vsp.DcInput) -> None:
        """# Write a DC analysis."""

        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Write the analysis command
        param = an.indep_name
        ## Interpret the sweep
        sweep_type = an.sweep.WhichOneof("tp")
        if sweep_type == "linear":
            sweep = an.sweep.linear
            line = f"{an.analysis_name} dc param={param} start={sweep.start} stop={sweep.stop} step={sweep.step}"
            self.writeln(line)
        elif sweep_type == "points":
            sweep = an.sweep.points
            line = f"{an.analysis_name} dc values=[{sweep.points}]"
            self.writeln(line)
        elif sweep_type == "log":
            raise NotImplementedError
        else:
            raise ValueError("Invalid sweep type")

    def write_op(self, an: vsp.OpInput) -> None:
        """# Write an operating point analysis."""

        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        self.writeln(f"{an.analysis_name} dc oppoint=rawfile")

    def write_tran(self, an: vsp.TranInput) -> None:
        """# Write a transient analysis."""

        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        if len(an.ctrls):
            raise NotImplementedError
        if len(an.ic):
            raise NotImplementedError

        self.writeln(f"{an.analysis_name} tran stop={an.tstop} ")

    def write_noise(self, an: vsp.NoiseInput) -> None:
        """# Write a noise analysis."""
        raise NotImplementedError



"""
# VlsirTools Pytest Utilities 

Primarily adds and parses the command-line options for enabling and disabling `vlsirtools.spice` simulators. 
To use these utilities, add the following to your `conftest.py`:

```python
from vlsirtools.pytest import (
    pytest_configure,
    pytest_addoption,
    pytest_collection_modifyitems,
)
```

See VlsirTools' own `conftest.py` for an example.  
With these functions included, by those names, in `conftest.py`, invocations of `pytest` can use: 

* The `--simulator_test_mode` command-line option 
  * Values are `required`, `if_available`, or `disabled`
* Markers for each supported simulator
  * `spectre`, `xyce`, and `ngspice` are currently supported
* Command-line options for each supported simulator, to override `simulator_test_mode`


Examples: 

```
pytest --simulator_test_mode disabled       # Disables all simulation tests
pytest --simulator_test_mode required       # Requires all simulation tests
pytest --simulator_test_mode if_available   # Runs tests on available simulators

pytest --simulator_test_mode if_available --ngspice required  # Requires ngspice, tests all others available 
pytest --simulator_test_mode required --xyce disabled         # Disables xyce, requires all others
pytest --simulator_test_mode disabled --spectre required      # Requires spectre, disables all others
```

"""


import pytest
from enum import Enum
from typing import Dict, Optional
from dataclasses import dataclass

import vlsirtools.spice as vsp


class SimulatorTestMode(Enum):
    """# Enumerated Per-Simulator Test Modes"""

    REQUIRED = "required"
    IF_AVAILABLE = "if_available"
    DISABLED = "disabled"

    def skip(self) -> bool:
        """Boolean indication of whether this Simulator should be skipped."""
        return self.value == "disabled"


@dataclass
class Simulator:
    marker: str
    # Simulator name/ command-line option/ pytest-marker name
    available: bool
    # Boolean indicator of availability in the current environment

    def getoption(self, config: "Config") -> Optional["SimulatorTestMode"]:
        """Get this simulator's pytest option"""
        option = config.getoption(f"--{self.marker}")
        if option is None:
            return None
        return SimulatorTestMode(option)


# The "registry" of supported simulators and their availability in the current process environment
# In other words, the stuff about the simulators that we know before we start running tests
simulators: Dict[str, Simulator] = {
    s.marker: s
    for s in [
        Simulator(marker="spectre", available=vsp.spectre.available()),
        Simulator(marker="xyce", available=vsp.xyce.available()),
        Simulator(marker="ngspice", available=vsp.ngspice.available()),
    ]
}


class Why(Enum):
    """# Enumerated reasons to be enabled or disabled."""

    DISABLED = "DISABLED"
    NOT_AVAILABLE = "NOT_AVAILABLE"
    AVAILABLE = "AVAILABLE"
    REQUIRED = "REQUIRED"

    def enabled(self) -> bool:
        """# Boolean enabledness indicator"""
        if self in (Why.DISABLED, Why.NOT_AVAILABLE):
            return False
        if self in (Why.REQUIRED, Why.AVAILABLE):
            return True
        raise ValueError(f"Unknown Why: {self}")


@dataclass
class SimulatorSettingPair:
    """# Simulator-Setting Pair
    The combination of a `Simulator` and its current test-setting."""

    simulator: Simulator  # The simulator
    mode: SimulatorTestMode  # Configured test mode
    enabled: bool  # Boolean indicator of whether this simulator is enabled
    why: Why  # Why this simulator is enabled or disabled

    @staticmethod
    def new(simulator: Simulator, mode: SimulatorTestMode) -> "SimulatorSettingPair":
        """# Create a SimulatorSettingPair from a Simulator and a test mode."""

        # Mostly suss out the reason "why"
        if mode == SimulatorTestMode.DISABLED:
            why = Why.DISABLED
        elif mode == SimulatorTestMode.IF_AVAILABLE:
            why = Why.AVAILABLE if simulator.available else Why.NOT_AVAILABLE
        elif mode == SimulatorTestMode.REQUIRED:
            why = Why.REQUIRED
        else:
            raise ValueError(f"Unknown mode: {mode}")

        return SimulatorSettingPair(
            simulator=simulator,
            mode=mode,
            enabled=why.enabled(),
            why=why,
        )


def pytest_configure(config):
    """# PyTest Configuration
    Adds a marker for each supported simulator."""

    for s in simulators.keys():
        config.addinivalue_line("markers", f"{s}")


def pytest_addoption(parser):
    """PyTest Command-Line Option Additions"""

    # Add the `--simulator_test_mode` option
    parser.addoption(
        "--simulator_test_mode",
        action="store",
        default=SimulatorTestMode.IF_AVAILABLE.value,
        help=f"Simulation test mode. One of {[m.value for m in SimulatorTestMode]}.",
    )
    # Add an option for each supported simulator
    for s in simulators.keys():
        parser.addoption(
            f"--{s}",
            action="store",
            default=None,
            help=f"{s.title()} test mode. One of {[m.value for m in SimulatorTestMode]}.",
        )


# Skip markers
skip_disabled = pytest.mark.skip(reason="Disabled by command-line simulator options")
skip_not_available = pytest.mark.skip(
    reason="Simulator not available & not required by command-line options"
)


def pytest_collection_modifyitems(config, items):
    """Examine each test item, and skip them based on `SimulatorSettings`."""

    # First grab the "overall mode"
    default_mode = SimulatorTestMode(config.getoption("--simulator_test_mode"))

    # Now check for any simulator-specific overrides, and turn them into setting-pairs
    simulator_settings = {
        s.marker: SimulatorSettingPair.new(
            simulator=s, mode=s.getoption(config) or default_mode
        )
        for s in simulators.values()
    }

    # print(f"Testing with simulation settings:")
    # print(simulator_settings)

    # Alright finally the point: apply the `skip` attribute to test we don't wanna run.
    for item in items:
        for marker in simulators.keys():
            if marker in item.keywords:
                pair = simulator_settings[marker]
                if pair.why == Why.DISABLED:
                    item.add_marker(skip_disabled)
                elif pair.why == Why.NOT_AVAILABLE:
                    item.add_marker(skip_not_available)



"""
NGSpice Implementation of `vlsir.spice.Sim`
"""

# Std-Lib Imports
from concurrent.futures import ProcessPoolExecutor
import subprocess, re, shutil
from enum import Enum
from warnings import warn
from dataclasses import dataclass
from typing import Mapping, IO, Dict
import shlex

# External Imports
import numpy as np

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from ..netlist import netlist
from ..netlist.spice import NgspiceNetlister
from .base import Sim
from .sim_data import TranResult, OpResult, SimResult, AcResult, DcResult, NoiseResult
from .spice import SupportedSimulators, sim

# Module-level configuration. Over-writeable by sufficiently motivated users.

# The simulator executable invoked. If over-ridden, likely for sake of a specific path or version.
NGSPICE_EXECUTABLE = "ngspice"


def available() -> bool:
    return NGSpiceSim.available()


class NGSpiceSim(Sim):
    """
    State and execution logic for a NGSpice-call to `vsp.Sim`.
    """

    @staticmethod
    def available() -> bool:
        """Boolean indication of whether the current running environment includes the simulator executable."""
        if shutil.which(NGSPICE_EXECUTABLE) is None:
            return False
        try:
            subprocess.run(
                f"{NGSPICE_EXECUTABLE} -v",
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=10,
            )
        except Exception:
            # Indicate "not available" for any Exception. Usually this will be a `subprocess.CalledProcessError`.
            return False
        return True  # Otherwise, installation looks good.

    @classmethod
    def enum(cls) -> SupportedSimulators:
        return SupportedSimulators.NGSPICE

    def run(self) -> SimResult:
        """Run the specified `SimInput` in directory `self.rundir`, returning its results."""

        # Write the netlist
        self.write_netlist()

        self.run_sim_process()

        # Handle stdout and stderr as needed
        # Parse up the results
        return self.parse_results()

    def write_netlist(self) -> None:
        """# Write our netlist to file"""

        netlist_file = self.open("netlist.sp", "w")
        netlister = NgspiceNetlister(dest=netlist_file)
        netlister.write_sim_input(self.inp)
        netlist_file.flush()
        netlist_file.close()

    def parse_results(self) -> SimResult:
        """# Parse output data"""

        data = parse_nutbin(self.open("netlist.raw", "rb"))
        an_type_dispatch = dict(
            ac=self.parse_ac,
            dc=self.parse_dc,
            op=self.parse_op,
            tran=self.parse_tran,
            noise=self.parse_noise,
        )
        an_name_dispatch = dict(
            op="Plotname: Operating Point\n",
            dc="Plotname: DC Analysis\n",
            ac="Plotname: AC Analysis\n",
            tran="Plotname: Transient Analysis\n",
            noise="FIXME! do we still want this setup?",
        )
        results = []
        for an in self.inp.an:
            an_type = an.WhichOneof("an")
            inner = getattr(an, an_type)

            if an_type == "noise":
                # FIXME: I read somewhere that "Special cases aren't special enough to break the rules."
                results.append(self.parse_noise(inner, data))
                continue

            if an_type not in an_type_dispatch:
                msg = f"Invalid or Unsupported analysis {an} with type {an_type}"
                raise RuntimeError(msg)
            func = an_type_dispatch[an_type]
            analysis_name = an_name_dispatch[an_type]
            if analysis_name not in data:
                msg = f"Cannot read results for analysis {an}"
                raise RuntimeError(msg)
            inner_data = data[analysis_name]
            an_results = func(inner, inner_data)
            results.append(an_results)

        return SimResult(an=results)

    def parse_ac(self, an: vsp.AcInput, nutbin: "NutBinAnalysis") -> AcResult:
        # FIXME: the `mt0` and friends file names collide with tran, if they are used in the same Sim!
        measurements = self.get_measurements("*.mt*")

        # Pop the frequence vector out of the data
        freq = nutbin.data.pop("frequency")

        # Nutbin format stores the frequency vector as complex numbers, along with all the complex-valued signal data.
        # NOTE: once upon a time, we checked that the imaginary part of all frequencies was zero.
        # And that used to work! Now it doesn't; ngspice just seemingly leaves it uninitialized, or set to some random value.
        # See https://github.com/Vlsir/Vlsir/issues/66
        # Now, just grab the real part.
        freq = freq.real

        return AcResult(
            analysis_name=an.analysis_name,
            freq=freq,
            data=nutbin.data,
            measurements=measurements,
        )

    def parse_dc(self, an: vsp.DcInput, nutbin: "NutBinAnalysis") -> DcResult:
        measurements = self.get_measurements("*.ms*")
        return DcResult(
            analysis_name=an.analysis_name,
            indep_name=an.indep_name,
            data=nutbin.data,
            measurements=measurements,
        )

    def parse_op(self, an: vsp.OpInput, nutbin: "NutBinAnalysis") -> OpResult:
        return OpResult(
            analysis_name=an.analysis_name,
            data={k: v[0] for k, v in nutbin.data.items()},
        )

    def parse_tran(self, an: vsp.TranInput, nutbin: "NutBinAnalysis") -> TranResult:
        """Extract the results for Analysis `an` from `data`."""
        measurements = self.get_measurements("*.mt*")
        return TranResult(
            analysis_name=an.analysis_name, data=nutbin.data, measurements=measurements
        )

    def parse_noise(
        self, an: vsp.NoiseInput, nutbin: Mapping[str, "NutBinAnalysis"]
    ) -> NoiseResult:
        # Noise results come in two "analyses": the frequency-sweep data, and the integrated noise.
        # Collate them into a `NoiseResult`
        # FIXME: add measurements (if there are such things?)
        noise_sweep_data = nutbin.get("Plotname: Noise Spectral Density Curves\n", None)
        if noise_sweep_data is None:
            raise RuntimeError
        integrated_noise = nutbin.get("Plotname: Integrated Noise\n", None)
        if integrated_noise is None:
            raise RuntimeError
        return NoiseResult(
            analysis_name=an.analysis_name,
            data=noise_sweep_data,
            integrated_noise=integrated_noise,
            measurements={},
        )

    def get_measurements(self, filepat: str) -> Dict[str, float]:
        """Get the measurements at files matching (glob) `filepat`.
        Returns only a single files-worth of measurements, and issues a warning if more than one such file exists.
        Returns an empty dictionary if no matching files are found."""
        meas_files = list(self.glob(filepat))
        if not meas_files:
            return dict()
        if len(meas_files) > 1:
            msg = f"Unsupported: more than one measurement-file generated. Only the first will be read"
            warn(msg)
        return parse_mt0(self.open(meas_files[0], "r"))

    def run_sim_process(self) -> None:
        """Run a NGSpice sub-process, executing the simulation"""
        # Note the `nutbin` output format is dictated here
        cmd = shlex.split(f"{NGSPICE_EXECUTABLE} -b netlist.sp -r netlist.raw")
        return self.run_subprocess(cmd)


class FromStr:
    """Mix-in for creating `Enum`s from string values."""

    @classmethod
    def from_str(cls, s: str) -> "FromStr":
        reversed = {v.value: v for v in cls}
        if s not in reversed:
            msg = f"Invalid `{cls.__name__}`: `{s}`"
            raise ValueError(msg)
        return reversed[s]


class NumType(FromStr, Enum):
    """# Numeric Datatypes
    Values equal the strings used in NutBin data files."""

    REAL = "real"
    COMPLEX = "complex"


class Units(FromStr, Enum):
    """# Unit Types
    Values equal the strings used in NutBin data files."""

    DUMMY = "dummy"
    SWEEP = "sweep"
    SECONDS = "seconds"
    VOLTS = "voltage"
    AMPS = "current"
    HERTZ = "frequency"
    CELSIUS = "C"

    @classmethod
    def from_str(cls, s: str) -> "FromStr":
        reversed_l = {v.value: v for v in cls}
        if s not in reversed_l:
            return cls.DUMMY
        return reversed_l[s]


@dataclass
class VarSpec:
    """Variable Spec"""

    name: str
    units: Units


@dataclass
class NutBinAnalysis:
    """Analysis result from a NutBin file."""

    analysis_name: str  # Analysis name
    numtype: NumType  # Numeric type in `data` field
    data: Mapping[str, np.ndarray]  # Signal name => data
    units: Mapping[str, Units]  # Signal name => units


def parse_nutbin(f: IO) -> Mapping[str, NutBinAnalysis]:
    """Parse a `nutbin` format set of simulation results.
    Returns results as a dictionary from `analysis_name` to `NutBinAnalysis`.
    Note this is paired with the simulator invocation commands, which include `format=nutbin`."""

    # Parse per-file header info
    # First 2 lines are ascii one line statements

    # And parse the rest of the file per-analysis
    rv = {}
    while f:
        _title = f.readline()  # Title, ignored
        _date = f.readline()  # Run date, ignored
        plotname = f.readline().decode("ascii")  # Simulation name
        if len(plotname) == 0:
            break
        an = parse_nutbin_analysis(f, plotname)
        rv[an.analysis_name] = an
    return rv


def parse_nutbin_analysis(f: IO, plotname: str) -> NutBinAnalysis:
    """Parse a `NutBinAnalysis` from an open nutbin-format file `f`."""

    # Next 4 lines are also ascii one line statements
    # Parse the `Flags` field, which primarily includes the numeric datatype
    flags = f.readline().decode("ascii")
    flags = flags.split()
    if len(flags) != 2 or flags[0] != "Flags:":
        raise ValueError(f"Invalid flags {flags}")
    numtype = NumType.from_str(flags[1])
    nptypes = {NumType.REAL: float, NumType.COMPLEX: complex}
    nptype = nptypes[numtype]

    # Parse the number of variables and points
    num_vars_line = f.readline()  # No. Variables:   [nvar]
    num_pts_line = f.readline()  # No. Points:      [npts]
    sim_name = plotname.split("`")[-1].split("'")[0]

    # Find the number of variables and number of points
    num_vars = int(
        re.match(
            r"No. Variables:\s+(?P<num_vars>\d+)\s*\n",
            num_vars_line.decode("ascii"),
        ).group("num_vars")
    )
    num_pts = int(
        re.match(
            r"No. Points:\s+(?P<num_pts>\d+)\s*\n",
            num_pts_line.decode("ascii"),
        ).group("num_pts")
    )

    # Decode the variables spec, looks like the following
    # Variables: [Variable idx] [Variable name] [units] [optional_flags]
    f.readline()
    var_specs = []
    for i in range(num_vars):
        var_specs.append(_read_var_spec(f.readline().decode("ascii")))

    # Read the binary data, should look like the following:
    # Binary: \n[Binary data]
    binary_line = f.readline().decode("ascii")
    assert binary_line == "Binary:\n"
    # Data is big endian
    bin_data = np.fromfile(
        f, dtype=np.dtype(nptype).newbyteorder("<"), count=num_vars * num_pts
    )
    data = {}
    units = {}
    for i, var in enumerate(var_specs):
        data[var.name] = bin_data[i::num_vars]
        units[var.name] = var.units

    return NutBinAnalysis(
        analysis_name=sim_name,
        numtype=numtype,
        data=data,
        units=units,
    )


def _read_var_spec(line: str) -> VarSpec:
    """Read a Variable spec line from the input
    and return the name and the units."""
    m = re.match(
        r"\s+(?P<idx>\d+)\s+(?P<name>\S+)\s+(?P<units>\S+)(?P<rest>.*)\n", line
    )
    return VarSpec(name=m.group("name"), units=Units.from_str(m.group("units")))


def parse_mt0(file: IO) -> Dict[str, float]:
    """Parse an (open) "mt0-format" measurement-file into a set of {name: value} pairs."""

    file.readline()  # Header
    file.readline()  # Netlist Title
    keys = file.readline()  # Measurement Names Line
    keys = keys.split()
    values = file.readline()  # Measurement Values Line

    def convert(s: str) -> float:
        """Convert a string to a float, converting failing cases to `NaN`"""
        try:
            return float(s)
        except:
            return float("NaN")

    values = [convert(s) for s in values.split()]
    return dict(zip(keys, values))



"""
# Vlsir Tools
"""

__version__ = "2.0.dev0"

# Python module namespaces
from . import spice
from . import spicetype
from .spicetype import SpiceType

# Primitive Definitions
from . import primitives  # The python module

# Note `vlsirtools.netlist` becomes a *function* here, where there is also a *module* by that name. Maybe not ideal.
from .netlist import netlist

# Pull the `spice/{simulator}` namespaces into the `vlsirtools` namespace as well
# FIXME: probably integrate netlisting into these at some point
from .spice import xyce
from .spice import spectre
from .spice import ngspice



"""
# Verilog-Format Netlister 
"""
# Local Imports
import vlsir
import vlsir.circuit_pb2 as vckt

# Import the base-class
from .base import Netlister, ResolvedModule, ResolvedParams


class VerilogNetlister(Netlister):
    """
    # Structural Verilog Netlister
    """

    @property
    def enum(self):
        """Get our entry in the `NetlistFormat` enumeration"""
        from . import NetlistFormat

        return NetlistFormat.VERILOG

    def write_module_definition(self, module: vckt.Module) -> None:
        """Create a Verilog module definition for proto-Module `module`"""

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Collect and index vckt.Signals in this Module by name.
        self.collect_signals_by_name(module)

        # Create the module header
        self.writeln(f"module {module_name}")

        # Write its parameters, if defined
        self.write_param_declarations(module)

        if module.ports:  # Create its ports
            # Don't forget, a trailing comma after the last one is fatal to high-tech Verilog parsers!
            self.writeln("( ")
            self.indent += 1
            for num, pport in enumerate(module.ports):
                comma = "" if num == len(module.ports) - 1 else ","
                self.writeln(self.format_port_decl(pport) + comma)
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("( ) ;  // No ports ")

        self.writeln("")  # Blank line to end "header" facets
        self.indent += 1

        # Write the internal signal declarations
        self.write_internal_signal_declarations()

        if module.instances:  # Create its instances
            self.writeln("")
            self.writeln("// Instance Declarations")
            for pinst in module.instances:
                self.write_instance(pinst)
        else:
            self.writeln("// No Instances")

        # Write any netlist-literal content
        self.write_literals(module.literals)

        # Close up the module
        self.indent -= 1
        self.writeln("")  # Blank before `endmodule`
        self.writeln(f"endmodule // {module_name} \n\n")

    def write_internal_signal_declarations(self) -> None:
        if not self.internal_signals_by_name:
            return self.writeln("// No Signal Declarations")

        self.writeln("// Signal Declarations")
        for psig in self.internal_signals_by_name.values():
            self.writeln(self.format_signal_decl(psig) + "; ")

    def write_param_declarations(self, module: vlsir.circuit.Module) -> None:
        """Write the parameter declarations for Module `module`."""

        if not len(module.parameters):  # No parameters
            # Note we check `len` because `module.parameters` is a protobuf thing, and *who knows* how it converts to `bool`.
            return self.writeln("// No parameters ")

        # Get all the formatted parameter-strings
        formatted = [self.format_param_decl(param) for param in module.parameters]
        # Add the commas to all *BUT THE LAST*. Like JSON, Verilog demands it, or fails.
        formatted = [d + "," for d in formatted[:-1]] + [formatted[-1]]

        # Write them all to our destination
        self.writeln("#( ")
        self.indent += 1
        [self.writeln(f) for f in formatted]
        self.indent -= 1
        self.writeln(") \n")

    def write_instance(self, pinst: vckt.Instance) -> None:
        """Format and write Instance `pinst`"""

        # Get its Module or ExternalModule definition
        rmodule = self.resolve_reference(pinst.module)
        if not isinstance(rmodule, ResolvedModule):
            # Spice-level primitives and models are generally not available in Verilog runtimes,
            # and hence generate errors here if attempted in netlisting.
            raise RuntimeError(f"Invalid module for Verilog: {rmodule}")
        module, module_name = rmodule.module, rmodule.module_name

        # Resolve its parameter values
        resolved_instance_parameters = self.get_instance_params(pinst, module)

        # Write the module-name
        self.writeln(module_name)

        # Write its parameter-values
        self.write_instance_params(resolved_instance_parameters)

        # Write the instance name
        self.writeln(pinst.name)

        if module.ports:  # Write connections, by-name, in-order
            self.writeln("( ")
            self.indent += 1
            # Get `module`'s port-order
            port_order = [pport.signal for pport in module.ports]
            connection_targets = {
                connection.portname: connection.target
                for connection in pinst.connections
            }
            # And write the Instance ports, in that order
            for num, pname in enumerate(port_order):
                ptarget = connection_targets.get(pname, None)
                if ptarget is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                # Again a trailing comma after the last one is fatal!
                comma = "" if num == len(port_order) - 1 else ","
                self.writeln(
                    f".{pname}({self.format_connection_target(ptarget)}){comma} "
                )
            # Close up the ports
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("// No ports ")

        self.writeln("")  # Post-Instance blank line

    def write_instance_params(self, pvals: ResolvedParams) -> None:
        """Write Instance parameters `pvals`"""
        if not pvals:
            return self.writeln("// No parameters ")

        # Write the parameter-values
        self.writeln("#( ")
        self.indent += 1
        # ANSI-style params: .NAME(VALUE)
        formatted = ", ".join([f".{pname}({pval})" for pname, pval in pvals.items()])
        self.writeln(formatted)
        self.indent -= 1
        self.writeln(") ")

    @classmethod
    def format_param_type(cls, pparam: vlsir.Param) -> str:
        """Verilog type-string for `Parameter` `param`."""
        ptype = pparam.WhichOneof("value")
        if ptype == "integer":
            return "longint"
        if ptype == "double":
            return "real"
        if ptype == "string":
            return "string"
        raise ValueError

    @classmethod
    def format_param_decl(cls, param: vlsir.Param) -> str:
        """Format a parameter-declaration"""
        rv = f"parameter {param.name}"
        # FIXME: whether to include datatype
        # dtype = cls.format_param_type(param)
        default = cls.get_param_default(param)
        if default is not None:
            rv += f" = {default}"
        return rv

    def format_concat(self, pconc: vckt.Concat) -> str:
        """Format the Concatenation of several other Connections"""
        # Verilog { a, b, c } concatenation format
        parts = [self.format_connection_target(part) for part in pconc.parts]
        return "{" + ", ".join(parts) + "}"

    def format_port_decl(self, pport: vckt.Port) -> str:
        """Format a `Port` declaration"""

        # First retrieve and check the validity of its direction
        port_type_to_str = {
            vckt.Port.Direction.Value("INPUT"): "input",
            vckt.Port.Direction.Value("OUTPUT"): "output",
            vckt.Port.Direction.Value("INOUT"): "inout",
            vckt.Port.Direction.Value("NONE"): "NO_DIRECTION",
        }
        dir_ = port_type_to_str.get(pport.direction, None)
        if dir_ is None:
            msg = f"Invalid Verilog netlisting for unknown Port direction {pport.direction}"
            raise RuntimeError(msg)
        if dir_ == "NO_DIRECTION":
            msg = f"Invalid Verilog netlisting for undirected Port {pport}"
            raise RuntimeError(msg)

        return dir_ + " " + self.format_signal_decl(self.get_signal(pport.signal))

    def format_signal_decl(self, psig: vckt.Signal) -> str:
        """Format a `Signal` declaration"""
        rv = "wire"
        if psig.width > 1:
            rv += f" [{psig.width-1}:0]"
        rv += f" {psig.name}"
        return rv

    def format_port_ref(self, pport: vckt.Port) -> str:
        """Format a reference to a `Port`.
        Unlike declarations, this just requires the name of its `Signal`."""
        return self.format_signal_ref(self.get_signal(pport.signal))

    @classmethod
    def format_signal_ref(cls, psig: vckt.Signal) -> str:
        """Format a reference to a `Signal`.
        Unlike declarations, this just requires its name."""
        return psig.name

    @classmethod
    def format_signal_slice(cls, pslice: vckt.Slice) -> str:
        """Format Signal-Slice `pslice`"""
        if pslice.top == pslice.bot:  # Single-bit slice
            return f"{pslice.signal}[{pslice.top}]"
        return f"{pslice.signal}[{pslice.top}:{pslice.bot}]"  # Multi-bit slice

    def write_comment(self, comment: str) -> None:
        """Verilog uses C-style line comments, beginning with `//`"""
        self.write(f"// {comment}\n")

    @classmethod
    def format_prefix(cls, pre: vlsir.SIPrefix) -> str:
        """Format a `SIPrefix` to a string"""
        # Verilog does not have the SI prefixes built in. Always write the exponent value.
        map = {
            # Single-character aliases, supported by every SPICE we know
            vlsir.SIPrefix.YOCTO: "e-24",
            vlsir.SIPrefix.ZEPTO: "e-21",
            vlsir.SIPrefix.ATTO: "e-18",
            vlsir.SIPrefix.FEMTO: "e-15",
            vlsir.SIPrefix.PICO: "e-12",
            vlsir.SIPrefix.NANO: "e-9",
            vlsir.SIPrefix.MICRO: "e-6",
            vlsir.SIPrefix.MILLI: "e-3",
            vlsir.SIPrefix.CENTI: "e-2",
            vlsir.SIPrefix.DECI: "e-1",
            vlsir.SIPrefix.UNIT: "",
            vlsir.SIPrefix.DECA: "e1",
            vlsir.SIPrefix.HECTO: "e2",
            vlsir.SIPrefix.KILO: "e3",
            vlsir.SIPrefix.MEGA: "e6",
            vlsir.SIPrefix.GIGA: "e9",
            vlsir.SIPrefix.TERA: "e12",
            vlsir.SIPrefix.PETA: "e15",
            vlsir.SIPrefix.EXA: "e17",
            vlsir.SIPrefix.ZETTA: "e18",
            vlsir.SIPrefix.YOTTA: "e19",
        }
        if pre not in map:
            raise ValueError(f"Invalid or Unsupported SIPrefix {pre}")

        return map[pre]



""" 
# Simulator State Base-Class 
"""

# Std-Lib Imports
import subprocess, os, tempfile
from typing import Optional, List, IO
from pathlib import Path

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
import vlsir.circuit_pb2 as vckt
from . import SimResultUnion, SupportedSimulators, SimOptions, ResultFormat


class Sim:
    """
    # Simulator State Base-Class

    Shared basic functionality for simulators, including:
    * Creating run-directories, and named files within them
    * Asynchronously invoking simulation subprocesses
    * Finding a valid top-level module
    * Exception/ Error types for common simulator errors

    Each `SupportedSimulator` is generally implemented as a sub-class of `Sim`.
    The interface between the `Sim` base-class and its implementing sub-class is such that:

    * The base-class `Sim.sim` class-method should really be the entry point for everything! It will:
      * Create the class instance
      * Set up the run-directory via `setup()`
      * Invoke the primary sub-class-specific method `run`
      * After completion, perform any clean-up activites via `cleanup()`.
    * All other methods are implemented by the sub-classes.
      * The only required method is `run`, which performs the actual simulation.
      * Other methods will generally do things including: writing simulator-specific netlist-content, launching a sim process, parsing results.
      * At no point should the sub-classes need to know any more about the `Sim` base-class, or call any of its `super` methods.
    """

    @classmethod
    def enum(cls) -> SupportedSimulators:
        raise NotImplementedError

    @classmethod
    def sim(
        cls, inp: vsp.SimInput, opts: Optional[SimOptions] = None
    ) -> SimResultUnion:
        """Sim-invoking class method.
        Creates an instance of `cls` as a context manager, run in its simulation directory.
        This should be invoked by typical implementations of a free-standing `sim` function."""

        if opts is None:  # Create the default `SimOptions`
            opts = SimOptions(simulator=cls.enum())

        # Create the simulation-class instance, and execute its main `run` method
        sim = cls(inp=inp, opts=opts)
        try:
            sim.setup()
            results = sim.run()
        finally:
            sim.cleanup()

        # FIXME: we shouldn't need this `isinstance`; get Xyce to return `sd.SimResult` and decide whether to convert here
        if opts.fmt == ResultFormat.VLSIR_PROTO and not isinstance(
            results, vsp.SimResult
        ):
            return results.to_proto()
        return results

    def run(self) -> SimResultUnion:
        raise NotImplementedError("`Sim` subclasses must implement `run`")

    def __init__(self, inp: vsp.SimInput, opts: SimOptions) -> None:
        self.inp = inp
        self.opts = opts
        self.rundir = opts.rundir
        self.tmpdir: Optional[tempfile.TemporaryDirectory] = None
        self.subprocesses: List[subprocess.Process] = []

    def setup(self):
        """Perform simulation setup, including the simulation directory and top-level Module validation."""

        # Set up the simulation directory
        if self.rundir is not None:  # User-specified `rundir`
            self.tmpdir = None
            self.rundir = Path(self.rundir).absolute()
            if not self.rundir.exists():
                os.makedirs(self.rundir)
        else:  # Create a new temp directory
            self.tmpdir = tempfile.TemporaryDirectory()
            self.rundir = Path(self.tmpdir.name).absolute()

    def cleanup(self):
        """On completion, clean up after ourselves."""
        if self.tmpdir is not None:
            self.tmpdir.cleanup()

    def run_subprocess(self, cmd: str) -> None:
        """Run a shell subprocess invoking command `cmd`.
        All subprocesses are run in `self.rundir`, and tracked in the list `self.subprocesses`."""

        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=str(self.rundir),
        )
        self.subprocesses.append(proc)
        stdout, stderr = proc.communicate()

        # The subprocess module does not raise Python exceptions: check the return code instead.
        if proc.returncode != 0:
            from . import SimError

            raise SimError(sim=self, stdout=stdout, stderr=stderr)
        return None

    def open(self, name: str, mode: str = "r") -> IO:
        """Open a file in the simulation directory."""
        return self.path(name).open(mode)

    def path(self, name: str) -> Path:
        """Return a path in the simulation directory."""
        return Path(self.rundir) / Path(name)

    def glob(self, pat: str):
        """Return a list of paths in the simulation directory matching `pat`."""
        return Path(self.rundir).glob(pat)



"""
# Netlister Base Class 
"""

# Std-Lib Imports
from typing import Optional, Union, IO, Dict, Iterable, List, Tuple
from dataclasses import dataclass, field

# Local Imports
import vlsir
import vlsir.circuit_pb2 as vckt
import vlsir.spice_pb2 as vsp
from .. import primitives
from ..spicetype import SpiceType

# Internal type shorthand
ModuleLike = Union[vckt.Module, vckt.ExternalModule]


@dataclass
class ResolvedModule:
    """Resolved reference to a `Module` or `ExternalModule`.
    Includes its spice-language prefix, and if user-defined its netlist-sanitized module-name.
    """

    module: ModuleLike
    module_name: str


@dataclass
class SpiceBuiltin:
    """# Reference to a SPICE built-in element, e.g. the ideal resistor, capacitor, or voltage source.
    Many formats include special syntax for defining these, e.g. parameter specifications that work for no other instance."""

    module: vckt.ExternalModule
    # External module definition, generally from `vlsir.primitives`. Largely used for ports and parameters.
    spice_type: SpiceType  # Spice type. One of the "non-model" variants.


@dataclass
class SpiceModelRef:
    """# Reference to a SPICE Model, e.g. a MOSFET model, defined with some variant of the `.model` statement."""

    module: vckt.ExternalModule  # External module definition
    model_name: str  # Model name
    spice_type: SpiceType  # Spice type. Can be any variant but SUBCKT.


# Union type of the targets for netlist instances
ResolvedRef = Union[ResolvedModule, SpiceBuiltin, SpiceModelRef]


@dataclass
class ResolvedParams:
    """Resolved Instance-Parameter Values
    Factoring in defaults, and converted to strings.
    Largely a wrapper for `Dict[str, str]`, with accessors `get` and `pop` that raise `RuntimeError` if a key is missing.
    """

    inner: Dict[str, str]

    def set(self, key: str, val: str) -> None:
        """Set the value of `key` to `val` in the resolved parameters."""
        self.inner[key] = val

    def get(self, key: str) -> str:
        """Get the value of `key` from the resolved parameters.
        Raises `RuntimeError` if `key` is not present."""
        if key not in self.inner:
            raise RuntimeError(f"Missing parameter {key}")
        return self.inner[key]

    def pop(self, key: str) -> str:
        """Get the value of `key` from the resolved parameters, and remove it from the `ResolvedParams`.
        Raises `RuntimeError` if `key` is not present."""
        if key not in self.inner:
            raise RuntimeError(f"Missing parameter {key}")
        return self.inner.pop(key)

    def pop_many(self, keys: Iterable[str]) -> Dict[str, str]:
        """Get the values of `keys` from the resolved parameters, and remove them from the `ResolvedParams`.
        Raises `RuntimeError` if any `key` is not present."""
        return {key: self.pop(key) for key in keys}

    def rename(self, old: str, new: str) -> None:
        """Rename a key from `old` to `new`.
        Raises an exception if `old` is not a key."""
        val = self.pop(old)
        self.set(new, val)

    @property
    def items(self):
        return self.inner.items

    def __bool__(self) -> bool:
        """Boolean conversions, generally through the `not` keyword or `bool` constructor,
        are forwarded down to the internal `inner` dictionary."""
        return bool(self.inner)

    def __contains__(self, key: str) -> bool:
        return key in self.inner


@dataclass
class Indent:
    """
    # Indentation Helper

    Supports in-place addition and subtraction of indentation levels, e.g. via
    ```python
    indent = Indent()
    indent += 1 # Adds one "tab", or indentation level
    indent += 1 # Adds another
    indent -= 1 # Drops back by one
    ```
    The current indentation-string is available via the `state` attribute.
    Writers using such an indenter will likely be of the form:
    ```python
    dest.write(f"{indent.state}{content}")
    ```
    """

    # Per-"tab" indentation characters. Defaults to two spaces.
    chars: str = 2 * " "
    # Current (integer) indentation-level, in number of "tabs"
    num: int = field(init=False, default=0)
    # Current indentation-string. Always equals `num * chars`.
    state: str = field(init=False, default="")

    def __post_init_post_parse__(self) -> None:
        self.state = self.chars * self.num

    def __iadd__(self, other: int) -> None:
        """In-place add, i.e. `indent += 1`"""
        self.num += other
        self.state = self.chars * self.num
        return self

    def __isub__(self, other: int) -> None:
        """In-place subtract, i.e. `indent -= 1`"""
        self.num = self.num - other
        if self.num < 0:
            raise ValueError("Negative indentation")
        self.state = self.chars * self.num
        return self


class Netlister:
    """# Abstract Base `Netlister` Class

    `Netlister` is not directly instantiable, and none of its sub-classes are intended
    for usage outside the `netlist` package. The primary API method `netlist` is designed to
    create, use, and drop a `Netlister` instance.
    Once instantiated a `Netlister`'s primary API method is `netlist`.
    This writes all content in its `pkg` field to destination `dest`.

    Internal methods come in three primary flavors:
    * `write_*` methods, which write to `self.dest`. These methods are generally format-specific.
    * `format_*` methods, which return format-specific strings, but *do not* write to `dest`.
    * `get_*` methods, which retrieve some internal data, e.g. extracting the type of a `Connection`.
    """

    def __init__(self, dest: IO):
        self.dest = dest
        self.indent = Indent(chars="  ")

        self.module_names = set()  # Netlisted Module names
        self.pmodules = dict()  # Visited proto-Modules

        # Keep two dictionaries tracking `ExternalModule`s:
        # * The VLSIR schema identifies them to two-string (domain, name) tuples. This is their "primary key".
        # * Most netlist formats use a single global namespace, and require a unique name for each `ExternalModule`.
        # So it's possible for two distinct VLSIR ExternalModules to conflict in netlist-language.
        # If and when this happens netlisting will raise an exception.
        # FIXME: that could maybe be configurable?
        self.ext_modules_by_key: Dict[Tuple[str, str], vckt.ExternalModule] = dict()
        self.ext_modules_by_name: Dict[str, vckt.ExternalModule] = dict()

        # Similar tracking of ExternalModules which resolve to SPICE `.model` definitions
        # Note in many formats, sub-circuits and models have different namespaces -
        # i.e. it is possible to have a sub-circuit and a model with the same name.
        self.spice_models_by_name: Dict[str, vckt.ExternalModule] = dict()

        # Attributes of the currently-netlisted Module

        # Signals in the currently-visited module, keyed by name
        # Includes *all* ports and internal signals
        self.signals_by_name: Dict[str, vckt.Signal] = dict()
        # Internal signals, keyed by name, *excluding* ports
        self.internal_signals_by_name: Dict[str, vckt.Signal] = dict()
        # Names of all ports, for membership testing
        self.port_names = set()  # : Set[str]

    """
    # Core Interactions with our Destination `IO`
    """

    def write(self, s: str) -> None:
        """Helper/wrapper, passing to `self.dest`"""
        self.dest.write(s)

    def writeln(self, s: str) -> None:
        """Write `s` as a line, at our current `indent` level."""
        self.write(f"{self.indent.state}{s}\n")

    def flush(self) -> None:
        """Flush `self.dest`."""
        self.dest.flush()

    """ 
    # Data Model Helpers
    """

    def get_external_module(self, emod: vckt.ExternalModule) -> None:
        """# Visit an `ExternalModule`
        "Netlisting" these doesn't actually write anything, but just stores a reference in internal dictionaries for future references to them.

        Note each netlister keeps two dictionaries tracking `ExternalModule`s:
        * The VLSIR schema identifies them to two-string (domain, name) tuples. This is their "primary key".
        * Most netlist formats use a single global namespace, and require a unique name for each `ExternalModule`.
        So it's possible for two distinct VLSIR ExternalModules to conflict in netlist-language.
        If and when this happens netlisting will raise an exception.
        FIXME: that could maybe be configurable?
        """

        # Check for duplicate definitions by "primary key" (domain, name)
        # Note we *do not* check on a name-only basis here; this is left to instantiation-time,
        # largely so that we can determine which ExternalModules resolve to subcircuits vs SPICE models.
        key = (emod.name.domain, emod.name.name)
        if key in self.ext_modules_by_key:
            raise RuntimeError(f"Invalid doubly-defined external module {emod}")
        self.ext_modules_by_key[key] = emod

    @classmethod
    def get_param_default(cls, pparam: vlsir.Param) -> Optional[str]:
        """Get the default value of `pparam`. Returns `None` for no default."""
        if pparam.value.WhichOneof("value") is None:
            return None
        return cls.get_param_value(pparam.value)

    @classmethod
    def get_param_value(cls, ppval: vlsir.ParamValue) -> str:
        """Get a string representation of a parameter-value"""
        ptype = ppval.WhichOneof("value")
        if ptype == "integer":
            return str(int(ppval.integer))
        if ptype == "double":
            return str(float(ppval.double))
        if ptype == "string":
            # String-valued parameters get embedded in double-quotes
            return f'"{str(ppval.string)}"'
        if ptype == "literal":
            # Whereas *literal* strings, e.g. "22.22e6-11e-1", do not.
            return str(ppval.literal)
        if ptype == "prefixed":
            return cls.format_prefixed(ppval.prefixed)
        raise ValueError(f"Invalid Param type {ptype}")

    @classmethod
    def get_instance_params(
        cls, pinst: vckt.Instance, pmodule: ModuleLike
    ) -> ResolvedParams:
        """Resolve the parameters of `pinst` to their values, including default values provided by `pmodule`.
        Raises a `RuntimeError` if any required parameter is not defined.

        Note this method *does not* raise errors for parameters *not specified* in `pmodule`,
        allowing for "pass-through" parameters not explicitly defined."""

        values = dict()
        # The price of not attaching this as a property of the instance is that we have to recompute it every time.
        instance_parameters = dict()
        for param in pinst.parameters:
            instance_parameters[param.name] = param.value

        # Step through each of `pmodule`'s declared parameters first, applying defaults if necessary
        for mparam in pmodule.parameters:
            if mparam.name in instance_parameters:  # Specified by the Instance
                inst_pval = instance_parameters.pop(mparam.name)
                values[mparam.name] = cls.get_param_value(inst_pval)
            else:  # Not specified by the instance. Apply the default, or fail.
                pdefault = cls.get_param_default(mparam)
                if pdefault is None:
                    msg = f"Required parameter `{mparam.name}` not specified for Instance `{pinst}`"
                    raise RuntimeError(msg)
                values[mparam.name] = pdefault

        # Convert the remaining instance-provided parameters to strings
        for pname, pval in instance_parameters.items():
            values[pname] = cls.get_param_value(pval)

        # And wrap the resolved values in a `ResolvedParams` object
        return ResolvedParams(values)

    @classmethod
    def get_module_name(cls, module: vckt.Module) -> str:
        """Create a netlist-compatible name for proto-Module `module`"""

        # Create the module name
        # Replace all format-invalid characters with underscores
        name = module.name.split(".")[-1]
        for ch in name:
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                name = name.replace(ch, "_")
        return name

    def resolve_reference(self, ref: vlsir.utils.Reference) -> ResolvedRef:
        """Resolve the `ModuleLike` referent of `ref`."""

        if ref.WhichOneof("to") == "local":  # Internally-defined Module
            module = self.pmodules.get(ref.local, None)
            if module is None:
                raise RuntimeError(f"Invalid undefined Module {ref.local} ")
            return ResolvedModule(
                module=module,
                module_name=self.get_module_name(module),
            )

        if ref.WhichOneof("to") == "external":  # Defined outside package
            # First check the priviledged/ internally-defined domains
            if ref.external.domain == "vlsir.primitives":
                # Built-in primitive. Load its definition from the `vlsir.primitives` (python) module.
                name = ref.external.name
                module = primitives.dct.get(ref.external.name, None)
                if module is None:
                    raise RuntimeError(f"Invalid undefined primitive {ref.external}")

                schema_spice_type = vckt.SpiceType.Name(module.spicetype)
                if schema_spice_type is None:
                    msg = f"Invalid SpiceType for {module}"
                    raise RuntimeError(msg)
                spice_type = SpiceType[schema_spice_type]

                # Double-check it has a valid SpiceType
                # Note the `vlsir.primitives` module *itself* should be well-behaved, only defining types in the `valid_types` list below.
                # But it's Python! Nothing stops *users* from modifying them, or trying to define their own!

                if spice_type == SpiceType.SUBCKT:
                    msg = f"Invalid SUBCKT in the `vlsir.primitives` namespace: {ref}\n"
                    msg += "(Did you try to define this yourself?)\n"
                    msg += "`vlsir.primitives` is a priviledged namespace, and cannot be modified."
                    raise RuntimeError(msg)

                model_based = {
                    SpiceType.MOS,
                    SpiceType.BIPOLAR,
                    SpiceType.DIODE,
                    SpiceType.TLINE,
                }
                if spice_type in model_based:
                    msg = f"Invalid/ deprecated model-based `vlsir.primitive` {ref}"
                    raise RuntimeError(msg)

                valid_types = {
                    SpiceType.RESISTOR,
                    SpiceType.CAPACITOR,
                    SpiceType.INDUCTOR,
                    SpiceType.VSOURCE,
                    SpiceType.ISOURCE,
                    SpiceType.VCVS,
                    SpiceType.VCCS,
                    SpiceType.CCCS,
                    SpiceType.CCVS,
                }
                if spice_type not in valid_types:
                    raise ValueError(f"Unsupported or Invalid Primitive {ref}")

                return SpiceBuiltin(
                    module=module,
                    spice_type=spice_type,
                )

            if ref.external.domain == "hdl21.primitives":
                msg = f"Invalid direct-netlisting of physical `hdl21.Primitive` `{ref.external.name}`. "
                msg += "Either compile to a target technology, or replace with an `ExternalModule`. "
                raise RuntimeError(msg)

            else:  # Externally-Defined, External-Domain `ExternalModule`
                key = (ref.external.domain, ref.external.name)
                module = self.ext_modules_by_key.get(key, None)
                if module is None:
                    msg = f"Invalid Instance of undefined External Module {key}"
                    raise RuntimeError(msg)

                schema_spice_type = vckt.SpiceType.Name(module.spicetype)
                if schema_spice_type is None:
                    msg = f"Invalid SpiceType for {module}"
                    raise RuntimeError(msg)
                spice_type = SpiceType[schema_spice_type]

                if spice_type == SpiceType.SUBCKT:  # External sub-circuit

                    # Check for duplicates on a name-only basis
                    # Most netlist formats have a single global namespace, so same-names conflict in netlist-language.
                    module_name = ref.external.name
                    cached = self.ext_modules_by_name.get(module_name, None)
                    if cached is not None and cached is not module:
                        msg = f"Conflicting ExternalModule definitions {module} and {cached}"
                        raise RuntimeError(msg)
                    self.ext_modules_by_name[module_name] = module

                    return ResolvedModule(
                        module=module,
                        module_name=module_name,
                    )

                else:  # SPICE Model Reference

                    # Again check for duplicates on a name-only basis
                    modelname = ref.external.name
                    cached = self.spice_models_by_name.get(modelname, None)
                    if cached is not None and cached is not module:
                        msg = f"Conflicting ExternalModule definitions {module} and {cached}"
                        raise RuntimeError(msg)
                    self.spice_models_by_name[modelname] = module

                    return SpiceModelRef(
                        module=module,
                        model_name=modelname,
                        spice_type=spice_type,
                    )

        # Not a Module, not an ExternalModule, not sure what it is
        raise ValueError(f"Invalid Module reference {ref}")

    def get_signal(self, name: str) -> vckt.Signal:
        """Get Signal `name` from the current Module's mapping.
        Raises a `RuntimeError` if the Signal is not found."""

        sig = self.signals_by_name.get(name, None)
        if sig is None:
            msg = f"Unknown signal: {name} in {self.signals_by_name.keys()}"
            raise RuntimeError(msg)
        return sig

    def collect_signals_by_name(self, module: vckt.Module):
        """Collect a `Module`'s worth of signals into a dictionary keyed by name.
        This often proves important for references to internal Signals, e.g. in Ports and Slices.
        """

        # Reset the state of our mappings
        self.signals_by_name = {}
        self.port_names = set()
        self.internal_signals_by_name = {}

        # Collect all the port-names into our set
        for port in module.ports:
            if port.signal in self.port_names:
                raise RuntimeError(f"Duplicate Port {port.signal}")
            self.port_names.add(port.signal)

        # Collect all Signals into dictionaries
        for signal in module.signals:
            if signal.name in self.signals_by_name:
                msg = f"Duplicate signal definition in Module {module.name}"
                raise RuntimeError(msg)
            self.signals_by_name[signal.name] = signal
            if signal.name not in self.port_names:
                self.internal_signals_by_name[signal.name] = signal

    @classmethod
    def validate_sim_top(cls, inp: vsp.SimInput) -> vckt.Module:
        """# Ensure that `SimInput` `inp`'s `top` module exists,
        and adheres to the "Spice top-level" port-interface: a single port for ground / VSS / node-zero.
        Returns the top-level `ModuleLike` when successful, or raises a `RuntimeError` when not.
        # FIXME: make `ExternalModule` valid too!
        """
        if not inp.top:
            cls.fail(f"No top-level module specified in {inp}")

        # Build a mapping from names to Modules
        module_name_dict: Dict[str, vckt.Module] = {m.name: m for m in inp.pkg.modules}

        # Check that the top-level module exists
        top = module_name_dict.get(inp.top, None)
        if top is None:
            msg = f"Top-level module `{inp.top}` not found among Modules {list(module_name_dict.keys())}"
            cls.fail(msg)

        # Check that the top-level module has a single port, for ground / VSS / node-zero
        if len(top.ports) != 1:
            msg = f"`vlsir.SimInput` top-level module {inp.top} must have *one* (VSS) port - has {len(module.ports)} ports [{module.ports}]"
            cls.fail(msg)

        return top

    """
    # Write Methods Implemented Here 
    Largely the things which are *not* format-specific, but instead traverse the data model and dispatch to other, format-specific methods.
    """

    def write_sim_input(self, inp: vsp.SimInput) -> None:
        """# Write simulation-input `inp` to `self.dest`."""

        # Write the header
        self.write_sim_header(inp)

        # Write the circuit-definitions package
        self.write_package(pkg=inp.pkg)

        # Write the top-level instance
        self.write_sim_dut(inp)

        # Write sim options
        self.write_sim_options(inp.opts)

        # Write each control element
        self.write_control_elements(inp.ctrls)

        # Write each analysis
        for an in inp.an:
            self.write_analysis(an)

        # Write a few blanks at the end
        self.write(3 * "\n")

        # And ensure all output makes it to `self.dest`
        self.dest.flush()

    def write_sim_header(self, inp: vsp.SimInput) -> None:
        """# Write header commentary for a `SimInput`
        This proves particularly important for many Spice-like formats,
        which *always* interpret the first line of an input-file as a comment (among many other dumb things).
        So, always write one there right off the bat."""

        self.write_comment(f"`{self.enum.value}` Sim Input for `{inp.top}`")
        self.write_comment(f"Generated by `vlsirtools.{self.__class__.__name__}`")
        self.write_comment("")

    def write_sim_dut(self, inp: vsp.SimInput) -> None:
        """# Write the top-level DUT instance for `inp`."""

        top = self.validate_sim_top(inp)
        top_name = self.get_module_name(top)
        self.writeln(self.format_sim_dut(top_name))

    def write_package(self, pkg: vckt.Package) -> None:
        """# Write circuit-Package `pkg` to `self.dest`."""

        # First visit any externally-defined Modules,
        # Ensuring we have their port-orders.
        for emod in pkg.ext_modules:
            self.get_external_module(emod)

        # Add some header commentary
        self.write_package_header(pkg)

        # Now do the real stuff,
        # Creating netlist entries for each package-defined Module
        for mod in pkg.modules:
            self.write_module_definition(mod)

        # And ensure all output makes it to `self.dest`
        self.dest.flush()

    def write_package_header(self, pkg: vckt.Package) -> None:
        """# Write header commentary for a `Package`
        This proves particularly important for many Spice-like formats,
        which *always* interpret the first line of an input-file as a comment (among many other dumb things).
        So, always write one there right off the bat."""

        if pkg.domain:
            self.write_comment(f"`circuit.Package` `{pkg.domain}`")
        else:
            self.write_comment(f"Anonymous `circuit.Package`")
        self.write_comment(f"Generated by `vlsirtools.{self.__class__.__name__}`")
        self.write_comment("")
        self.writeln("")

    def write_control_elements(self, ctrls: List[vsp.Control]) -> None:
        """# Write a list of `Control` elements"""
        for ctrl in ctrls:
            self.write_control_element(ctrl=ctrl)

    def write_control_element(self, ctrl: vsp.Control) -> None:
        """# Write a `Control` element"""

        # `Control` is largely a type-union of these things; dispatch across them.
        inner = ctrl.WhichOneof("ctrl")
        if inner == "include":
            return self.write_include(ctrl.include)
        if inner == "lib":
            return self.write_lib_include(ctrl.lib)
        if inner == "literal":
            return self.write_literal(ctrl.literal)
        if inner == "param":
            return self.write_sim_param(ctrl.param)
        if inner == "meas":
            return self.write_meas(ctrl.meas)
        if inner == "save":
            return self.write_save(ctrl.save)
        self.fail(f"Unknown control type: {inner}")

    def write_sim_options(self, options: List[vsp.SimOptions]) -> None:
        # FIXME: make this just `List[Param]` instead
        for opt in options:
            self.write_sim_option(opt)

    def write_analysis(self, an: vsp.Analysis) -> None:
        """# Write an `Analysis`, largely dispatching its content to a type-specific method."""

        inner = an.WhichOneof("an")
        inner_dispatch = dict(
            ac=self.write_ac,
            dc=self.write_dc,
            op=self.write_op,
            tran=self.write_tran,
            noise=self.write_noise,
        )
        if inner not in inner_dispatch:
            self.fail(f"Invalid analysis type {inner}")
        analysis_writer = inner_dispatch[inner]
        return analysis_writer(getattr(an, inner))

    def write_literal(self, literal: str) -> None:
        """# Write a literal string"""
        return self.writeln(literal)

    """ 
    # Format Methods Implemented Here
    """

    def format_connection_target(self, ptarget: vckt.ConnectionTarget) -> str:
        """Format a `ConnectionTarget` reference.
        Does not *declare* any new connection objects, but generates references to existing ones.
        """
        # `ConnectionTarget`s are a proto `oneof` union
        # which includes signals, slices, and concatenations.
        # Figure out which to import

        stype = ptarget.WhichOneof("stype")
        if stype == "sig":
            signal = self.get_signal(ptarget.sig)
            return self.format_signal_ref(signal)
        if stype == "slice":
            return self.format_signal_slice(ptarget.slice)
        if stype == "concat":
            return self.format_concat(ptarget.concat)
        raise ValueError(f"Invalid Type {stype} for Connection Target {ptarget}")

    @classmethod
    def format_prefixed(cls, pre: vlsir.Prefixed) -> str:
        prefix = cls.format_prefix(pre.prefix)
        numtp = pre.WhichOneof("number")
        if numtp == "integer":
            num = str(pre.integer)
        elif numtp == "string":
            num = str(pre.string)
        elif numtp == "double":
            raise ValueError(f"Deprecated double-valued Prefixed parameter {pre}")
        else:
            raise ValueError(f"Invalid `Prefixed` number type {numtp}")

        # Note the "prefix" is in fact at the *end*,
        # e.g. "5u", "11K".
        # (Calling it a *pre*-fix refers to *units*, not to numeric values)
        return f"{num}{prefix}"

    def write_literals(self, literals: List[str]) -> None:
        """# Write a list of literal strings, one per line."""
        for literal in literals:
            self.writeln(literal)

    """ 
    Virtual `write` Methods 
    """

    def write_comment(self, comment: str) -> None:
        """Format and string a string comment.
        "Line comments" are the sole supported variety, which fit within a line, and extend to the end of that line.
        The `write_comment` method assumes responsibility for closing the line."""
        raise NotImplementedError

    def write_module_definition(self, pmodule: vckt.Module) -> None:
        """Write Module `module`"""
        raise NotImplementedError

    def write_param_declarations(self, module: vckt.Module) -> None:
        """Write all parameter declarations for `module`"""
        raise NotImplementedError

    def write_instance(self, pinst: vckt.Instance) -> str:
        """Write Instance `pinst`"""
        raise NotImplementedError

    def write_instance_params(self, pvals: ResolvedParams) -> None:
        """Write Instance parameters `pvals`"""
        raise NotImplementedError

    def write_include(self, inc: vsp.Include) -> None:
        """# Write an `Include` statement"""
        raise NotImplementedError

    def write_lib_include(self, lib: vsp.LibInclude) -> None:
        """# Write a `LibInclude` statement"""
        raise NotImplementedError

    def write_save(self, save: vsp.Save) -> None:
        """# Write a `Save` statement"""
        raise NotImplementedError

    def write_meas(self, meas: vsp.Meas) -> None:
        """# Write a `Meas` statement"""
        raise NotImplementedError

    def write_sim_param(self, param: vlsir.Param) -> None:
        """# Write a simulation-level parameter"""
        raise NotImplementedError

    def write_sim_option(self, opt: vsp.SimOptions) -> None:
        """# Write a simulation option"""
        # FIXME: make this just `Param` instead
        raise NotImplementedError

    def write_ac(self, an: vsp.AcInput) -> None:
        """# Write an AC analysis."""
        raise NotImplementedError

    def write_dc(self, an: vsp.DcInput) -> None:
        """# Write a DC analysis."""
        raise NotImplementedError

    def write_op(self, an: vsp.OpInput) -> None:
        """# Write an operating point analysis."""
        raise NotImplementedError

    def write_tran(self, an: vsp.TranInput) -> None:
        """# Write a transient analysis."""
        raise NotImplementedError

    def write_noise(self, an: vsp.NoiseInput) -> None:
        """# Write a noise analysis."""
        raise NotImplementedError

    """ 
    Virtual `format` Methods 
    """

    @classmethod
    def format_param_decl(cls, param: vlsir.Param) -> str:
        """Format a named `Parameter` definition"""
        raise NotImplementedError

    @classmethod
    def format_port_decl(cls, pport: vckt.Port) -> str:
        """Format a declaration of a `Port`"""
        raise NotImplementedError

    @classmethod
    def format_port_ref(cls, pport: vckt.Port) -> str:
        """Format a reference to a `Port`"""
        raise NotImplementedError

    @classmethod
    def format_signal_decl(cls, psig: vckt.Signal) -> str:
        """Format a declaration of Signal `psig`"""
        raise NotImplementedError

    @classmethod
    def format_signal_ref(cls, psig: vckt.Signal) -> str:
        """Format a reference to Signal `psig`"""
        raise NotImplementedError

    @classmethod
    def format_signal_slice(cls, pslice: vckt.Slice) -> str:
        """Format Signal-Slice `pslice`"""
        raise NotImplementedError

    def format_concat(self, pconc: vckt.Concat) -> str:
        """Format the Concatenation of several other Connections"""
        raise NotImplementedError

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """Format bus-bit `index`"""
        raise NotImplementedError

    @classmethod
    def format_prefix(cls, pre: vlsir.SIPrefix) -> str:
        """Format a `SIPrefix` to a string"""
        raise NotImplementedError

    @classmethod
    def format_sim_dut(cls, module_name: str) -> str:
        """# Format the top-level DUT instance for module name `module_name`."""
        raise NotImplementedError

    """ 
    Other Virtual Methods 
    """

    @property
    def enum(self):
        """Get our entry in the `NetlistFormat` enumeration"""
        raise NotImplementedError

    """ 
    Other Helper Methods
    """

    @staticmethod
    def fail(msg: str) -> None:
        """# Error Helper. Raise a `RuntimeError` with message `msg`.
        Primarily for tracking nested state upon failure."""
        # Note: also a great place for a debugger breakpoint.
        raise RuntimeError(msg)



""" 
# Spice-Class Simulator Interface 
"""

# Std-Lib Imports
import os, subprocess
import concurrent.futures
from typing import Union, Optional, Sequence, TypeVar
from enum import Enum
from textwrap import dedent
from dataclasses import dataclass, field

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from vlsir.spice_pb2 import *  # Not used here intentionally, but "re-exported"
from . import sim_data as sd


class ResultFormat(Enum):
    """Enumerated Result Formats"""

    SIM_DATA = "sim_data"  # Types defined in the `sim_data` module
    VLSIR_PROTO = "vlsir_proto"  # `vsp.SimResults` and related protobuf-defined types


# Union of the two result-types, returned by many simulation methods
SimResultUnion = Union[vsp.SimResult, sd.SimResult]


class SupportedSimulators(Enum):
    """Enumerated, Internally-Defined Spice-Class Simulators"""

    SPECTRE = "spectre"
    XYCE = "xyce"
    NGSPICE = "ngspice"


def default() -> Optional[SupportedSimulators]:
    """Get the default simulator, for this Python-process and its environment.
    This largely consists of a priority-ordered walk through `SupportedSimulators`,
    returning the first whose `available()` method indicates its availability.
    Returns `None` if no such simulator appears available."""

    from .spectre import available as spectre_available
    from .xyce import available as xyce_available
    from .ngspice import available as ngspice_available

    if spectre_available():
        return SupportedSimulators.SPECTRE
    if xyce_available():
        return SupportedSimulators.XYCE
    if ngspice_available():
        return SupportedSimulators.NGSPICE
    return None  # Nothing found


@dataclass
class SimOptions:
    """Options to `vsp.Sim` which are *not* passed along to the simulator.
    I.e. options which effect the Python-process invoking simulation,
    but not the internals of the simulation itself."""

    # Simulator. FIXME: debatable whether to include this.
    simulator: SupportedSimulators = field(default_factory=default)

    # In-memory results format
    fmt: ResultFormat = ResultFormat.VLSIR_PROTO

    # Simulation run-directory. Uses a `tempdir` if unspecified.
    rundir: Optional[os.PathLike] = None


# Shorthand type alias for "an element or list thereof", used by all the call signatures below
T = TypeVar("T")
OneOrMore = Union[T, Sequence[T]]


def sim(
    inp: OneOrMore[vsp.SimInput], opts: Optional[SimOptions] = None
) -> OneOrMore[SimResultUnion]:
    """
    Concurrently execute one or more `vlir.spice.Sim`.
    Dispatches across `SupportedSimulators` specified in `SimOptions` `opts`.
    Uses the default `Simulator` as detected by the `default` method if no `simulator` is specified.
    """
    from .xyce import XyceSim
    from .spectre import SpectreSim
    from .ngspice import NGSpiceSim

    # Sort out the difference between "One" "OrMore" cases of input
    # For a single `SimInput`, create a list, but note we only want to return a single `SimResult`
    inp_is_a_single_sim = False
    if not isinstance(inp, Sequence):
        inp = [inp]
        inp_is_a_single_sim = True
    for x in inp:
        if not isinstance(x, vsp.SimInput):
            raise TypeError(f"Expected `vsp.SimInput`, got {x}")

    if opts is None:  # Create the default `SimOptions`
        opts = SimOptions()

    if opts.simulator is None:  # If we didn't specify or find a simulator, fail.
        msg = f"vlsirtools.spice: No Supported Simulators available for call to `sim()`"
        raise RuntimeError(msg)

    # Get the per-simulator callable
    if opts.simulator == SupportedSimulators.XYCE:
        cls = XyceSim
    elif opts.simulator == SupportedSimulators.SPECTRE:
        cls = SpectreSim
    elif opts.simulator == SupportedSimulators.NGSPICE:
        cls = NGSpiceSim
    else:
        raise ValueError(f"Unsupported simulator: {opts.simulator}")

    if len(inp) > 1 and opts.rundir is not None:
        # FIXME: how to ultimately handle this "multi-inputs plus specified directory" case
        raise RuntimeError("Cannot specify a run-directory for multiple simulations")

    # And do the real work, invoking the target simulator
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = [executor.submit(cls.sim, i, opts) for i in inp]
        results = [
            future.result() for future in concurrent.futures.as_completed(futures)
        ]

    # For the sequence of inputs case, return the sequence of results that came back
    if not inp_is_a_single_sim:
        return results

    # Unpack the single-input case
    if len(results) != 1:
        raise RuntimeError("Expected a single result")
    return results[0]


class SimError(Exception):
    """Exception raised when a simulation fails."""

    def __init__(self, sim: "Sim", stdout: bytes, stderr: bytes) -> None:
        s = dedent(
            f"""
            rundir: {str(sim.rundir)}
            stdout: {str(stdout)}
            stderr: {str(stderr)}
        """
        )
        super().__init__(s)
        self.rundir = sim.rundir
        self.stdout = stdout
        self.stderr = stderr


class SimProcessError(SimError):
    """Exception raised when an external simulator process fails."""

    def __init__(self, sim: "Sim", e: subprocess.CalledProcessError) -> None:
        super().__init__(sim=sim, stdout=e.stdout, stderr=e.stderr)



"""
# Sim Data 

A mirror of Spice Proto's Analysis Result objects for in-Python usage, 
using python data classes and numpy arrays.

Also provides round-tripping utilities between the two.
TODO: Go from proto -> sim_result
"""


from enum import Enum
from dataclasses import dataclass, field
from typing import List, Mapping, Union, ClassVar, Dict, Optional

import numpy as np

import vlsir.spice_pb2 as vsp


class AnalysisType(Enum):
    """Enumerated Analysis-Types
    Values are equal to the keys of the VLSIR `Analysis` type-union."""

    OP = "op"
    DC = "dc"
    AC = "ac"
    TRAN = "tran"
    NOISE = "noise"
    MONTE = "monte"
    SWEEP = "sweep"
    CUSTOM = "custom"


@dataclass
class OpResult:
    analysis_name: str
    data: Mapping[str, float]
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.OP

    def to_proto(self) -> vsp.OpResult:
        res = vsp.OpResult(analysis_name=self.analysis_name)
        for k, v in self.data.items():
            res.signals.append(k)
            res.data.append(v)
        return res


@dataclass
class DcResult:
    analysis_name: str
    indep_name: str
    data: Mapping[str, np.ndarray]
    measurements: Mapping[str, float]
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.DC

    def to_proto(self) -> vsp.DcResult:
        res = vsp.DcResult(analysis_name=self.analysis_name, indep_name=self.indep_name)
        for k, v in self.data.items():
            res.signals.append(k)
            for d in v:
                res.data.append(d)
        # TODO Add support for measurements
        return res


@dataclass
class TranResult:
    analysis_name: str
    data: Mapping[str, np.ndarray]
    measurements: Mapping[str, float]
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.TRAN

    def to_proto(self) -> vsp.TranResult:
        res = vsp.TranResult(analysis_name=self.analysis_name)
        for k, v in self.data.items():
            res.signals.append(k)
            for d in v:
                res.data.append(d)
        # TODO Add support for measurements
        return res


@dataclass
class AcResult:
    analysis_name: str  # Analysis name
    freq: np.ndarray  # Real/ float-valued frequency data
    data: Mapping[str, np.ndarray]  # Complex-valued signal data
    measurements: Mapping[str, float]  # Measurement data
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.AC

    def to_proto(self) -> vsp.AcResult:
        """Convert to a VLSIR `AcResult` proto object
        Primarily "flattens" the complex-valued data into a single list."""

        data: List[vsp.ComplexNum] = []
        for v in self.data.values():
            data.extend([vsp.ComplexNum(re=c.real, im=c.imag) for c in v])

        return vsp.AcResult(
            analysis_name=self.analysis_name,
            freq=self.freq,
            signals=list(self.data.keys()),
            data=data,
            measurements=self.measurements,
        )


@dataclass
class NoiseResult:
    analysis_name: str
    data: Mapping[str, np.ndarray]
    integrated_noise: Mapping[str, float]
    measurements: Mapping[str, float]
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.NOISE

    def to_proto(self) -> vsp.AcResult:
        raise NotImplementedError


@dataclass
class SweepResult:
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.SWEEP

    def __post_init__(self):
        raise NotImplementedError


@dataclass
class MonteResult:
    vlsir_type: ClassVar[AnalysisType] = AnalysisType.MONTE

    def __post_init__(self):
        raise NotImplementedError


@dataclass
class CustomAnalysisResult:
    """Custom Analysis "Results"
    No data is returned from these simulator-specific analyses;
    an empty `CustomAnalysisResult` is simply added to their overall results-list
    to pair with its input analysis, keeping all others aligned."""

    vlsir_type: ClassVar[AnalysisType] = AnalysisType.CUSTOM


# Type alias for the union of each result-type
AnalysisResult = Union[
    AcResult,
    DcResult,
    TranResult,
    OpResult,
    SweepResult,
    MonteResult,
    CustomAnalysisResult,
]


# Type union for indexing into AnalysisResults
AnalysisIndex = Union[int, str, AnalysisType, vsp.Analysis]


@dataclass
class AnalysisResultsTable:
    """# Multi-Indexed Table of Analysis Results
    Accessible by analysis-type, index, and name."""

    order: List[AnalysisResult]
    # In-order list of results, as ordered in `SimInput`.
    names: Dict[str, AnalysisResult]
    # Mapping by analysis-name
    types: Dict[AnalysisType, List[AnalysisResult]]
    # Mapping by analysis-type. Note there can be more than one of each type.

    @classmethod
    def create(cls, results: "SimResult") -> "AnalysisResultsTable":
        """Create an index from a `SimResult` object."""
        order = results.an
        names = {r.analysis_name: r for r in order}
        # "By type" is different, since there can be several of each type.
        types = {}
        for r in order:
            if r.vlsir_type in types:
                types[r.vlsir_type].append(r)
            else:
                types[r.vlsir_type] = [r]
        return cls(order, names, types)

    def get(self, key: AnalysisIndex) -> AnalysisResult:
        """Index into our results by type, index, or name."""

        if isinstance(key, int):
            return self.order[key]
        if isinstance(key, str):
            return self.names[key]
        if isinstance(key, AnalysisType):
            if key not in self.types:
                raise ValueError(f"No results for AnalysisType {key}")
            if len(self.types[key]) != 1:
                raise RuntimeError(f"Multiple results of type {key}")
            return self.types[key][0]
        raise TypeError(f"Invalid index into SimResults: {key} ({type(key)})")


@dataclass
class SimResult:
    """Results from a Mult-Analysis `Sim`"""

    an: List[AnalysisResult] = field(default_factory=list)
    # List of per-analysis results, in the same order specified in `SimInput.an`.

    def __post_init__(self):
        # Our multi-indexed table of results.
        # Unset until accessed with `get or __getitem__`
        self._table: Optional[AnalysisResultsTable] = None

    def index(self):
        """Recreate our multi-index table. Generally required on changes to primary list `an`."""
        self._table = AnalysisResultsTable.create(self)

    def __getitem__(self, key: AnalysisIndex) -> AnalysisResult:
        """Index into our results by type, index, or name."""
        return self.get(key)

    def get(self, key: AnalysisIndex) -> AnalysisResult:
        """Get one of our analysis-results by type, index, or name."""
        if self._table is None:
            self.index()
        return self._table.get(key)

    def to_proto(self) -> vsp.SimResult:
        """Convert to a VLSIR `SimResult` proto object."""
        res = vsp.SimResult()
        for an in self.an:
            ar = vsp.AnalysisResult(**{an.vlsir_type.value: an.to_proto()})
            res.an.append(ar)
        return res



"""
Xyce Implementation of `vsp.Sim`
"""

# Std-Lib Imports
import subprocess, random, shutil
import concurrent.futures
from glob import glob
from os import PathLike
from typing import IO, Dict, Union
import shlex

import numpy as np
import pandas as pd

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from ..netlist import XyceNetlister
from .base import Sim
from .sim_data import (
    TranResult,
    OpResult,
    SimResult,
    AcResult,
    DcResult,
    AnalysisResult,
)
from .spice import SupportedSimulators, sim


# Module-level configuration. Over-writeable by sufficiently motivated users.
XYCE_EXECUTABLE = "Xyce"  # The simulator executable invoked. If over-ridden, likely for sake of a specific path or version.


def available() -> bool:
    return XyceSim.available()


class XyceSim(Sim):
    """
    State and execution logic for a Xyce-call to `vsp.Sim`.

    Xyce can, in principle, run multiple analyses per process,
    but seems to commonly confuse outputs or disallow saving them
    from multiple analyses.

    Execution therefore instead occurs one Xyce-process per analysis.
    Results from each analysis-process are collated into a single `SimResult`.
    """

    @staticmethod
    def available() -> bool:
        """Boolean indication of whether the current running environment includes the simulator executable on its path."""

        if shutil.which(XYCE_EXECUTABLE) is None:
            return False
        try:
            # And if it's set, check that we can get its version without croaking.
            subprocess.run(
                f"{XYCE_EXECUTABLE} -v",  # Get the version,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=10,
            )
        except Exception:
            # Indicate "not available" for any Exception. Usually this will be a `subprocess.CalledProcessError`.
            return False
        return True  # Otherwise, installation looks good.

    @classmethod
    def enum(cls) -> SupportedSimulators:
        return SupportedSimulators.XYCE

    def run(self) -> SimResult:
        """Run the specified `SimInput` in directory `self.rundir`, returning its results."""

        # Write the DUT netlist
        self.write_dut_netlist()

        # Run each analysis as a concurrent subprocess
        with concurrent.futures.ThreadPoolExecutor() as executor:
            an_results = list(executor.map(self.analysis, self.inp.an))
        return SimResult(an_results)

    def write_dut_netlist(self) -> None:
        """# Write the DUT part (really, everything but Analyses) of the netlist.

        Xyce is weird in that we can't reliably get multi-analysis sims to work,
        so we write the DUT, copy it to a new file for each analysis, and then carry on modifying that.

        This function largely parallels `Netlister.write_sim_input()`,
        but skips the analysis-specific parts."""

        netlist_file = self.open("dut", "w")
        netlister = XyceNetlister(dest=netlist_file)

        # Do our fake version of `write_sim_input`

        # Write the header
        netlister.write_sim_header(self.inp)

        # Write the circuit-definitions package
        netlister.write_package(pkg=self.inp.pkg)

        # Write the top-level instance
        netlister.write_sim_dut(self.inp)

        # Write sim options
        netlister.write_sim_options(self.inp.opts)

        # Write each control element
        netlister.write_control_elements(self.inp.ctrls)

        # Skip this part!
        # # Write each analysis
        # for an in self.inp.an:
        #     netlister.write_analysis(an)

        # And ensure all output makes it to `self.dest`
        netlister.flush()

    def analysis(self, an: vsp.Analysis) -> AnalysisResult:
        """Execute a `vsp.Analysis`, returning its `AnalysisResult`"""

        # `Analysis` is a Union (protobuf `oneof`) of the analysis-types.
        # Unwrap it, and dispatch based on the type.
        inner = an.WhichOneof("an")

        if inner == "op":
            return self.op(an.op)
        if inner == "dc":
            return self.dc(an.dc)
        if inner == "ac":
            return self.ac(an.ac)
        if inner == "tran":
            return self.tran(an.tran)
        if inner in ("sweep", "monte", "custom"):
            raise NotImplementedError(f"{inner} not implemented")
        raise RuntimeError(f"Unknown analysis type: {inner}")

    def ac(self, an: vsp.AcInput) -> vsp.AcResult:
        """Run an AC analysis."""

        # Unpack the `AcInput`
        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        analysis_name = an.analysis_name
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        netlist = self.copy_dut_netlist(f"{analysis_name}.sp")

        # Write the analysis command
        npts = an.npts
        fstart = an.fstart
        fstop = an.fstop
        netlist.write(f".ac DEC {npts} {fstart} {fstop} \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print ac format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with self.open(f"{analysis_name}.sp.FD.csv", "r") as csv_handle:
            csv_data = read_csv(csv_handle)

        # Separate Frequency vector
        freq: np.ndarray = csv_data.pop("FREQ")
        data: Dict[str, np.ndarray] = {}

        # Pull together separate real/imaginary parts into complex numbers
        keys = list(csv_data.keys())
        for re, im in zip(keys[::2], keys[1::2]):
            # Check that node-names match, or fail.
            # Peel out "Re(" and "Im(" from the beginning, and a trailing ")" from the end.
            if re[3:-1] != im[3:-1]:
                raise RuntimeError(f"Unmatched complex number: {re}, {im}")
            name = re[3:-1]
            data[name] = np.array(
                [complex(r, i) for r, i in zip(csv_data[re], csv_data[im])]
            )

        # Parse any scalar measurement results
        measurements = self.parse_measurements(analysis_name)

        # And arrange them in an `AcResult`
        return AcResult(
            analysis_name=an.analysis_name,
            freq=freq,
            data=data,
            measurements=measurements,
        )

    def dc(self, an: vsp.DcInput) -> DcResult:
        """Run a DC analysis."""

        # Unpack the `DcInput`
        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        analysis_name = an.analysis_name
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        netlist = self.copy_dut_netlist(f"{analysis_name}.sp")

        # Write the analysis command
        param = an.indep_name
        ## Interpret the sweep
        sweep_type = an.sweep.WhichOneof("tp")
        if sweep_type == "linear":
            sweep = an.sweep.linear
            netlist.write(
                f".dc LIN {param} {sweep.start} {sweep.stop} {sweep.step}\n\n"
            )
        elif sweep_type == "log":
            sweep = an.sweep.log
            netlist.write(
                f".dc DEC {param} {sweep.start} {sweep.stop} {sweep.npts}\n\n"
            )
        elif sweep_type == "points":
            sweep = an.sweep.points
            netlist.write(
                f".dc {param} LIST {' '.join([str(pt) for pt in sweep.points])}\n\n"
            )
        else:
            raise ValueError("Invalid sweep type")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print dc format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with self.open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            csv_data = read_csv(csv_handle)

        # Parse any scalar measurement results
        measurements = self.parse_measurements(analysis_name)

        # And arrange them in an `OpResult`
        return DcResult(
            analysis_name=an.analysis_name,
            indep_name=an.indep_name,
            data=csv_data,
            measurements=measurements,
        )

    def op(self, an: vsp.OpInput) -> OpResult:
        """Run an operating-point analysis.
        Xyce describes the `.op` analysis as "partially supported".
        Here the `vsp.Op` analysis is mapped to DC, with a dummy sweep."""

        # Unpack the `OpInput`
        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        analysis_name = an.analysis_name
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        netlist = self.copy_dut_netlist(f"{analysis_name}.sp")

        # Create the dummy parameter, and "sweep" a single value of it
        dummy_param = f"_dummy_{random.randint(0,65536)}_"
        netlist.write(f".param {dummy_param}=1 \n\n")

        # Write the analysis command
        netlist.write(f".dc {dummy_param} 1 1 1 \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print dc format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with self.open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            csv_data = read_csv(csv_handle)

        # Each value in `csv_data` will be a single-element list.
        # Pull those single elements out.
        data = {k: v[0] for k, v in csv_data.items()}

        # And arrange them in an `OpResult`
        return OpResult(
            analysis_name=an.analysis_name,
            data=data,
        )

    def tran(self, an: vsp.TranInput) -> vsp.TranResult:
        """Run a transient analysis."""

        # Extract fields from our `TranInput`
        if not an.analysis_name:
            raise RuntimeError(f"Analysis name required for {an}")
        analysis_name = an.analysis_name

        # Why not make tstop/tstep required?
        if not an.tstop or not an.tstep:
            raise ValueError("tstop and tstep must be defined")
        tstop = an.tstop
        tstep = an.tstep
        if len(an.ic):
            raise NotImplementedError
        if len(an.ctrls):
            raise NotImplementedError

        # Copy and append to the existing DUT netlist
        netlist = self.copy_dut_netlist(f"{analysis_name}.sp")

        # Write the analysis command
        netlist.write(f".tran {tstep} {tstop} \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print tran format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Parse and organize our results
        # First pull them in from CSV
        with self.open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            csv_data = read_csv(csv_handle)

        # Parse any scalar measurement results
        measurements = self.parse_measurements(analysis_name)

        # And organize them into a `TranResult`
        return TranResult(
            analysis_name=an.analysis_name, data=csv_data, measurements=measurements
        )

    def run_xyce_process(self, name: str) -> None:
        """Run a `Xyce` sub-process executing the simulation."""
        return self.run_subprocess(cmd=shlex.split(f"{XYCE_EXECUTABLE} {name}.sp"))

    def parse_measurements(self, analysis_name: str) -> Dict[str, float]:
        # FIXME: the *input* should really be dictating whether we have measurements.
        # For now, we just search for any matching filenames via `glob`
        meas_glob = glob(f"*{analysis_name}*.m*0")
        if len(meas_glob) > 1:
            raise RuntimeError(f"Unsupported multiple measurement results for {self}")
        if len(meas_glob) == 1:
            measurements = parse_meas(self.open(meas_glob[0], "r"))
            return {k.lower(): v for k, v in measurements.items()}
        # No measurement-file, return an empty result
        return {}

    def copy_dut_netlist(self, path: str) -> IO:
        """Copy the `DUT` part of the netlist to file `path`,
        in our working directory, and return a file-handle to it."""
        shutil.copy(self.path("dut"), self.path(path))
        return self.open(path, "a")


def read_csv(handle: Union[IO, PathLike]) -> Dict[str, np.ndarray]:
    """Read CSV from file-handle `handle` into a dictionary of {header: array}s."""

    df = pd.read_csv(handle)
    return {n: np.array(c) for n, c in df.items()}


def parse_meas(file: IO) -> Dict[str, float]:
    """Parse an (open) measurement-file to a {name: value} dictionary."""
    rv = {}
    for line in file.readlines():
        contents = line.split()
        if len(contents) != 3 or contents[1] != "=":
            raise RuntimeError(f"Invalid line in Xyce measurements: {line}")
        name, val = contents[0], float(contents[2])
        rv[name] = val
    return rv



"""
# Vlsir Python Tools
Unit Tests
"""

import numpy as np
import pytest
from io import StringIO
from typing import Dict, List, Optional

import vlsir.utils_pb2 as vutils
from vlsir.utils_pb2 import Reference, QualifiedName, ParamValue, Param
import vlsir.circuit_pb2 as vckt
from vlsir.circuit_pb2 import (
    Module,
    Signal,
    Connection,
    ConnectionTarget,
    Port,
    Instance,
    Package,
)
import vlsir.spice_pb2 as vsp

import vlsirtools
from vlsirtools.spice import (
    SupportedSimulators,
    SimOptions,
    ResultFormat,
    sim,
)
import vlsirtools.spice.sim_data as sd
from vlsirtools.spice.sim_data import AnalysisType


"""
## Utility & Helper Functions
"""


def _connections(**kwargs):
    """Create a list of `Connection`s from keyword args of the form `portname=conn_target`, where `conn_target` is a `ConnectionTarget`."""
    return [Connection(portname=key, target=value) for key, value in kwargs.items()]


def _params(**kwargs):
    """Create a list of `Param`s from keyword args of the form `r=ParamValue(double=1e3)`"""
    return [Param(name=key, value=value) for key, value in kwargs.items()]


def _prim(name: str) -> Reference:
    """Create a `Reference` to primitive `name`"""
    return Reference(external=QualifiedName(domain="vlsir.primitives", name=name))


def test_version():
    assert vlsirtools.__version__ == "2.0.dev0"


def test_verilog_netlist1():
    """Test netlisting to a handful of formats, including Verilog-compatible contents."""

    # "Verilog Compatibility" requires:
    # * All ports must be directed. No "NONE" directions.
    # * No primitive instances.
    #
    # The test here covers module, signal, and port creation,
    # and basic instantiation of other local Modules.

    def _ports() -> List[vckt.Port]:
        return [
            Port(direction="INPUT", signal="inp"),
            Port(direction="OUTPUT", signal="out"),
            Port(direction="INOUT", signal="io"),
        ]

    def _signals() -> List[vckt.Signal]:
        return [
            Signal(name="inp", width=1),
            Signal(name="out", width=1),
            Signal(name="io", width=1),
        ]

    pkg = Package(
        domain="vlsirtools.tests.test_verilog_netlist1",
        modules=[
            # Inner, content-less Module, with a port of each direction
            Module(
                name="inner",
                parameters=_params(
                    a=ParamValue(integer=3),
                    b=ParamValue(double=1e-9),
                    d=ParamValue(literal="1+1"),
                    e=ParamValue(
                        prefixed=vutils.Prefixed(prefix="MICRO", string="11.11")
                    ),
                ),
                ports=_ports(),
                signals=_signals(),
            ),
            # Outer top Module which instantiates `inner`
            Module(
                name="top",
                ports=[],
                signals=_signals(),
                instances=[
                    Instance(
                        name="inner",
                        module=Reference(local="inner"),
                        parameters=_params(
                            a=ParamValue(integer=4),
                            b=ParamValue(double=2e-9),
                            d=ParamValue(literal="2+2"),
                            e=ParamValue(
                                prefixed=vutils.Prefixed(prefix="MICRO", string="22.22")
                            ),
                        ),
                        connections=[
                            Connection(portname=name, target=ConnectionTarget(sig=name))
                            for name in ("inp", "out", "io")
                        ],
                    )
                ],
            ),
        ],
    )
    dest = StringIO()
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="verilog")
    # While verilog is the point here, the other formats should work too:
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spice")
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spectre")
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="xyce")


def test_spice_netlist1():
    """Test spice netlisting, including instantiating primitives"""

    def default_conns() -> Dict:
        # Shorthand for connections between node "vvv" and "VSS", used for many instances below.
        return _connections(
            p=ConnectionTarget(sig="vvv"),
            n=ConnectionTarget(sig="VSS"),
        )

    # Create ExternalModules for a few model-based device-types
    mos1 = vlsirtools.primitives.mos(name="mos1", domain="mymodels")
    d1 = vlsirtools.primitives.diode(name="d1", domain="mymodels")
    bjt1 = vlsirtools.primitives.bipolar(name="bjt1", domain="mymodels")
    t1 = vlsirtools.primitives.tline(name="t1", domain="mymodels")

    def _eref(emod: vckt.ExternalModule) -> Reference:
        return Reference(external=emod.name)

    pkg = Package(
        domain="vlsir.tests.test_netlist1",
        ext_modules=[mos1, d1, bjt1, t1],
        modules=[
            Module(
                name="mid",
                parameters=_params(
                    r=ParamValue(double=1e3),
                    l=ParamValue(double=1e-9),
                    c=ParamValue(double=1e-15),
                ),
                ports=[
                    Port(direction="NONE", signal="vvv"),
                    Port(direction="NONE", signal="VSS"),
                ],
                signals=[Signal(name="vvv", width=1), Signal(name="VSS", width=1)],
                instances=[
                    Instance(
                        name="r",
                        module=_prim("resistor"),
                        connections=default_conns(),
                        parameters=_params(r=ParamValue(double=1e3)),
                    ),
                    Instance(
                        name="l",
                        module=_prim("inductor"),
                        connections=default_conns(),
                        parameters=_params(l=ParamValue(double=1e-9)),
                    ),
                    Instance(
                        name="c",
                        module=_prim("capacitor"),
                        connections=default_conns(),
                        parameters=_params(c=ParamValue(double=1e-15)),
                    ),
                    Instance(
                        name="v",
                        module=_prim("vdc"),
                        connections=default_conns(),
                        parameters=_params(
                            dc=ParamValue(double=1.1), ac=ParamValue(double=0)
                        ),
                    ),
                    Instance(
                        name="i",
                        module=_prim("isource"),
                        connections=default_conns(),
                        parameters=_params(dc=ParamValue(double=1e-6)),
                    ),
                    # Model-based instances
                    Instance(
                        name="m",
                        module=_eref(mos1),
                        connections=_connections(
                            d=ConnectionTarget(sig="VSS"),
                            g=ConnectionTarget(sig="VSS"),
                            s=ConnectionTarget(sig="VSS"),
                            b=ConnectionTarget(sig="VSS"),
                        ),
                        parameters=[],
                    ),
                    Instance(
                        name="q",
                        module=_eref(bjt1),
                        connections=_connections(
                            c=ConnectionTarget(sig="VSS"),
                            b=ConnectionTarget(sig="VSS"),
                            e=ConnectionTarget(sig="VSS"),
                        ),
                        parameters=[],
                    ),
                    Instance(
                        name="d",
                        module=_eref(d1),
                        connections=default_conns(),
                        parameters=[],
                    ),
                    Instance(
                        name="t",
                        module=_eref(t1),
                        connections=_connections(
                            p1p=ConnectionTarget(sig="VSS"),
                            p1n=ConnectionTarget(sig="VSS"),
                            p2p=ConnectionTarget(sig="VSS"),
                            p2n=ConnectionTarget(sig="VSS"),
                        ),
                        parameters=[],
                    ),
                ],
            ),
            Module(
                name="top",
                ports=[
                    Port(direction="NONE", signal="VSS"),
                ],
                signals=[Signal(name="vvv", width=1), Signal(name="VSS", width=1)],
                instances=[
                    Instance(
                        name="imid",
                        module=Reference(local="mid"),
                        connections=_connections(
                            vvv=ConnectionTarget(sig="vvv"),
                            VSS=ConnectionTarget(sig="VSS"),
                        ),
                    )
                ],
            ),
        ],
    )
    dest = StringIO()
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spice")


def test_netlist_hdl21_ideal1():
    """Test that netlisting a (deprecated) `hdl21.ideal` element fails with a `RuntimeError`."""

    pkg = Package(
        domain="vlsir.tests.test_netlist_hdl21_ideal1",
        modules=[
            Module(
                name="mid",
                ports=[
                    Port(direction="NONE", signal="vvv"),
                    Port(direction="NONE", signal="VSS"),
                ],
                signals=[Signal(name="vvv", width=1), Signal(name="VSS", width=1)],
                instances=[
                    Instance(
                        name="r1",
                        module=Reference(
                            external=QualifiedName(
                                domain="hdl21.ideal",
                                name="IdealResistor",  # FIXME: being deprecated
                            )
                        ),
                        connections=[
                            Connection(
                                portname="p", target=ConnectionTarget(sig="vvv")
                            ),
                            Connection(
                                portname="n", target=ConnectionTarget(sig="VSS")
                            ),
                        ],
                    ),
                ],
            )
        ],
    )
    dest = StringIO()
    with pytest.raises(RuntimeError):
        vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spice")


def dummy_testbench_package():
    """Create and return a `circuit.Package` with a single, (near) empty testbench module.
    Some simulators *really* don't like empty DUT content, and others don't like singly-connected nodes.
    So the simplest test-bench is two resistors, in parallel, between ground and a single "other node"."""

    def _r(name: str) -> Instance:
        # Create a canned instance of `vlsir.primitives.resistor` with instance-name `name`
        return Instance(
            name=name,  # <= Instance name argument here
            module=Reference(
                external=QualifiedName(domain="vlsir.primitives", name="resistor")
            ),
            connections=[
                Connection(portname="p", target=ConnectionTarget(sig="the_other_node")),
                Connection(portname="n", target=ConnectionTarget(sig="VSS")),
            ],
            parameters=[Param(name="r", value=ParamValue(double=1e3))],
        )

    return Package(
        domain="vlsirtools.tests.dummy_testbench_package",
        modules=[
            Module(
                name="dummy_testbench",
                ports=[
                    Port(direction="NONE", signal="VSS"),
                ],
                signals=[
                    Signal(name="VSS", width=1),
                    Signal(name="the_other_node", width=1),
                ],
                instances=[
                    _r("r1"),
                    _r("r2"),
                ],
            )
        ],
    )


def test_netlist_dummy_testbench():
    """Test netlisting the empty testbench package, used later for simulation tests"""

    dest = StringIO()
    vlsirtools.netlist(pkg=dummy_testbench_package(), dest=dest, fmt="spice")
    vlsirtools.netlist(pkg=dummy_testbench_package(), dest=dest, fmt="spectre")
    vlsirtools.netlist(pkg=dummy_testbench_package(), dest=dest, fmt="xyce")

    # The testbench package is not verilog-compatible; check that it fails.
    with pytest.raises(RuntimeError):
        vlsirtools.netlist(pkg=dummy_testbench_package(), dest=dest, fmt="verilog")


def dummy_sim(skip: Optional[List[AnalysisType]] = None):
    """Create a dummy `SimInput` for the `dummy_testbench` package."""

    if skip is None:
        skip = []

    # Set up an `analysis` for each analysis-type to run on the empty testbench.
    # FIXME: add noise, sweep, etc.
    analyses_by_type: Dict[AnalysisType, vsp.Analysis] = {
        AnalysisType.OP: vsp.Analysis(
            op=vsp.OpInput(
                analysis_name="op1",
                ctrls=[],
            )
        ),
        AnalysisType.DC: vsp.Analysis(
            dc=vsp.DcInput(
                analysis_name="dc1",
                indep_name="DUMMY",
                sweep=vsp.Sweep(
                    linear=vsp.LinearSweep(
                        start=0.0,
                        stop=1.0,
                        step=0.1,
                    ),
                ),
                ctrls=[],
            )
        ),
        AnalysisType.TRAN: vsp.Analysis(
            tran=vsp.TranInput(
                analysis_name="tr1",
                tstop=1e-9,
                tstep=1e-12,
                ic={},
                ctrls=[],
            )
        ),
        AnalysisType.AC: vsp.Analysis(
            ac=vsp.AcInput(
                analysis_name="ac1",
                fstart=1e3,
                fstop=1e6,
                npts=10,
                ctrls=[],
            )
        ),
    }
    # Rejigger those into a list, skipping any that are in the `skip` list.
    an = [a for t, a in analyses_by_type.items() if t not in skip]

    return vsp.SimInput(
        top="dummy_testbench",
        pkg=dummy_testbench_package(),
        an=an,
        ctrls=[
            vsp.Control(param=Param(name="DUMMY", value=ParamValue(integer=0))),
            # This parameter "DUMMY" is the sweep-variable for the DC analysis above.
        ],
    )


def dummy_sim_tests(
    simulator: SupportedSimulators, skip: Optional[List[AnalysisType]] = None
) -> sd.SimResult:
    """Run the `dummy_sim` input for simulator `simulator`.
    Returns the `sim_data.SimResult` object for any further inspection."""
    if skip is None:
        skip = []

    # Create the "dummy" `SimInput`
    inp = dummy_sim(skip=skip)

    # Run it, requesting in-memory `SimData` results.
    sd_results = sim(
        inp=inp,
        opts=SimOptions(
            simulator=simulator,
            fmt=ResultFormat.SIM_DATA,
            rundir="./scratch",
        ),
    )

    # Check a handful of things about what comes back
    assert isinstance(sd_results, sd.SimResult)
    if AnalysisType.OP not in skip:
        assert isinstance(sd_results[AnalysisType.OP], sd.OpResult)
    if AnalysisType.DC not in skip:
        assert isinstance(sd_results[AnalysisType.DC], sd.DcResult)
    if AnalysisType.TRAN not in skip:
        assert isinstance(sd_results[AnalysisType.TRAN], sd.TranResult)
    if AnalysisType.AC not in skip:
        assert isinstance(sd_results[AnalysisType.AC], sd.AcResult)
    converted_results = sd_results.to_proto()
    assert isinstance(converted_results, vsp.SimResult)

    # Run it again, but requesting VLSIR ProtoBuf output.
    proto_results = sim(
        inp=inp,
        opts=SimOptions(
            simulator=simulator,
            fmt=ResultFormat.VLSIR_PROTO,
        ),
    )
    assert isinstance(proto_results, vsp.SimResult)

    # And return the `sim_data` version for any further inspection.
    return sd_results


@pytest.mark.spectre
def test_spectre1():
    """Test an empty-input call to the `vlsir.spice.Sim` interface to `spectre`."""
    dummy_sim_tests(SupportedSimulators.SPECTRE)


@pytest.mark.xyce
def test_xyce1():
    """Test an empty-input call to the `vlsir.spice.Sim` interface to `xyce`."""
    dummy_sim_tests(SupportedSimulators.XYCE)


@pytest.mark.ngspice
def test_ngspice1():
    """Test an empty-input call to the `vlsir.spice.Sim` interface to `ngspice`."""

    res = dummy_sim_tests(
        SupportedSimulators.NGSPICE,
        skip=[
            AnalysisType.DC
        ],  ## Skip the DC sweep; ngspice doesn't support this kinda sweep
    )

    # Check that the AC frequency vector came back as expected
    # A follow-on from https://github.com/Vlsir/Vlsir/issues/66
    freqs = np.array(
        [
            1.00000000e03,
            1.25892541e03,
            1.58489319e03,
            1.99526231e03,
            2.51188643e03,
            3.16227766e03,
            3.98107171e03,
            5.01187234e03,
            6.30957344e03,
            7.94328235e03,
            1.00000000e04,
            1.25892541e04,
            1.58489319e04,
            1.99526231e04,
            2.51188643e04,
            3.16227766e04,
            3.98107171e04,
            5.01187234e04,
            6.30957344e04,
            7.94328235e04,
            1.00000000e05,
            1.25892541e05,
            1.58489319e05,
            1.99526231e05,
            2.51188643e05,
            3.16227766e05,
            3.98107171e05,
            5.01187234e05,
            6.30957344e05,
            7.94328235e05,
            1.00000000e06,
        ]
    )
    ac_result = res[AnalysisType.AC]
    assert np.allclose(ac_result.freq, freqs)


@pytest.mark.xyce
def test_xyce_import():
    # Just test importing from the `vlsirtools.xyce` path
    # FIXME: probably deprecate this
    from vlsirtools.xyce import sim


@pytest.mark.spectre
def test_spectre_import():
    # Just test importing from the `vlsirtools.spectre` path
    # FIXME: probably deprecate this
    from vlsirtools.spectre import sim


@pytest.mark.ngspice
def test_noise1():
    """Test the Noise analysis"""

    # A very complicated testbench: a voltage source and resistor in parallel.
    pkg = vckt.Package(
        domain="vlsirtools.tests.test_noise1",
        modules=[
            vckt.Module(
                name="noisetb",
                ports=[
                    vckt.Port(direction="NONE", signal="VSS"),
                ],
                signals=[
                    vckt.Signal(name="VSS", width=1),
                    vckt.Signal(name="the_other_node", width=1),
                ],
                instances=[
                    Instance(
                        name="r1",
                        module=Reference(
                            external=QualifiedName(
                                domain="vlsir.primitives", name="resistor"
                            )
                        ),
                        connections=[
                            Connection(
                                portname="p",
                                target=ConnectionTarget(sig="the_other_node"),
                            ),
                            Connection(
                                portname="n", target=ConnectionTarget(sig="VSS")
                            ),
                        ],
                        parameters=[Param(name="r", value=ParamValue(double=1e3))],
                    ),
                    Instance(
                        name="v1",
                        module=Reference(
                            external=QualifiedName(
                                domain="vlsir.primitives", name="vdc"
                            )
                        ),
                        connections=[
                            Connection(
                                portname="p",
                                target=ConnectionTarget(sig="the_other_node"),
                            ),
                            Connection(
                                portname="n", target=ConnectionTarget(sig="VSS")
                            ),
                        ],
                        parameters=[
                            Param(name="dc", value=ParamValue(double=0)),
                            Param(name="ac", value=ParamValue(double=0)),
                        ],
                    ),
                ],
            )
        ],
    )
    sim_input = vsp.SimInput(
        pkg=pkg,
        top="noisetb",
        an=[
            vsp.Analysis(
                noise=vsp.NoiseInput(
                    analysis_name="noise1",
                    output_p="the_other_node",
                    output_n="VSS",
                    input_source="v1",
                    fstart=1,
                    fstop=1e10,
                    npts=10,
                    ctrls=[],
                )
            )
        ],
    )
    vlsirtools.spice.sim(
        sim_input,
        opts=SimOptions(
            simulator=SupportedSimulators.NGSPICE, fmt=ResultFormat.SIM_DATA
        ),
    )



import vlsir
from .base import Netlister


class SpectreSpiceShared(Netlister):
    """Shared logic between Spectre and Spice netlisters.
    Designed to be a parent class of both."""

    @classmethod
    def format_param_decl(cls, param: vlsir.Param) -> str:
        """Format a parameter-declaration. The `value` field serves as the default value."""
        default = cls.get_param_default(param)
        if default is None:
            # FIXME: this may not be a requirement for *every* Spice netlist syntax.
            # But it's been a requirement throughout for ones that VLSIR supports.
            # At minimum, add some docs to this effect.
            msg = f"Invalid non-default parameter {param} for Spice netlisting"
            raise RuntimeError(msg)
        return f"{param.name}={default}"

    @classmethod
    def format_prefix(cls, pre: vlsir.SIPrefix) -> str:
        """Format a `SIPrefix` to a string"""
        # Use the single-character string where we can, and the exponent otherwise.
        map = {
            # Single-character aliases, supported by every SPICE we know
            vlsir.SIPrefix.ATTO: "a",
            vlsir.SIPrefix.FEMTO: "f",
            vlsir.SIPrefix.PICO: "p",
            vlsir.SIPrefix.NANO: "n",
            vlsir.SIPrefix.MICRO: "u",
            vlsir.SIPrefix.MILLI: "m",
            vlsir.SIPrefix.UNIT: "",
            vlsir.SIPrefix.KILO: "K",
            vlsir.SIPrefix.MEGA: "M",
            vlsir.SIPrefix.GIGA: "G",
            vlsir.SIPrefix.TERA: "T",
            vlsir.SIPrefix.PETA: "P",
            # Fall back to the exponent for the rest
            vlsir.SIPrefix.YOCTO: "e-24",
            vlsir.SIPrefix.ZEPTO: "e-21",
            vlsir.SIPrefix.CENTI: "e-2",
            vlsir.SIPrefix.DECI: "e-1",
            vlsir.SIPrefix.DECA: "e1",
            vlsir.SIPrefix.HECTO: "e2",
            vlsir.SIPrefix.EXA: "e17",
            vlsir.SIPrefix.ZETTA: "e18",
            vlsir.SIPrefix.YOTTA: "e19",
        }
        if pre not in map:
            raise ValueError(f"Invalid or Unsupported SIPrefix {pre}")

        return map[pre]



""" 
# Vlsir Netlisting 

Exports `vlsir.circuit.Package` to a netlist format.
"""

# Std-Lib Imports
from typing import Union, IO, Optional
from enum import Enum
from dataclasses import dataclass

# Local Imports
import vlsir

# Locally-defined netlister classes
from .spectre import SpectreNetlister
from .verilog import VerilogNetlister
from .spice import (
    SpiceNetlister,
    HspiceNetlister,
    CdlNetlister,
    XyceNetlister,
    NgspiceNetlister,
)


class NetlistFormat(Enum):
    """Enumeration of available formats.
    Includes string-value conversion."""

    VERILOG = "verilog"
    SPECTRE = "spectre"

    # Spice Dialects
    SPICE = "spice"
    NGSPICE = "ngspice"
    XYCE = "xyce"
    HSPICE = "hspice"
    CDL = "cdl"

    @staticmethod
    def get(spec: "NetlistFormatSpec") -> "NetlistFormat":
        """Get the format specified by `spec`, in either enum or string terms.
        Only does real work in the case when `spec` is a string, otherwise returns it unchanged."""
        if isinstance(spec, (NetlistFormat, str)):
            return NetlistFormat(spec)
        raise TypeError

    def netlister(self) -> type:
        """Get the paired netlister-class"""
        if self == NetlistFormat.SPECTRE:
            return SpectreNetlister
        if self == NetlistFormat.VERILOG:
            return VerilogNetlister
        if self == NetlistFormat.SPICE:
            return SpiceNetlister
        if self == NetlistFormat.HSPICE:
            return HspiceNetlister
        if self == NetlistFormat.NGSPICE:
            return NgspiceNetlister
        if self == NetlistFormat.XYCE:
            return XyceNetlister
        if self == NetlistFormat.CDL:
            return CdlNetlister
        raise ValueError


@dataclass
class NetlistOptions:
    """Options for netlisting."""

    indent: str = 2 * " "  # Indentation. Defaults to two spaces.
    width: int = 80  # Line-width. Defaults to 80.


# Type-alias for specifying format, either in enum or string terms
NetlistFormatSpec = Union[NetlistFormat, str]

## FIXME: add more `Netlistable`s
##Netlistable = Union[vlsir.circuit.Package]
Netlistable = vlsir.circuit.Package


def netlist(
    pkg: Netlistable,  ## FIXME: rename
    dest: IO,
    fmt: NetlistFormatSpec = "spectre",
    opts: Optional[NetlistOptions] = None,
) -> None:
    """Netlist proto-Package `pkg` to destination `dest`.

    Example usages:
    ```python
    h.netlist(pkg, dest=open('mynetlist.v', 'w'), fmt='verilog')
    ```
    ```python
    s = StringIO()
    h.netlist(pkg, dest=s, fmt='spectre')
    ```
    ```python
    import sys
    h.netlist(pkg, dest=sys.stdout, fmt='spice')
    ```

    Primary argument `pkg` must be a `vlsir.circuit.Package`.
    Destination `dest` may be anything that supports the `typing.IO` bundle,
    commonly including open file-handles. `StringIO` is particularly helpful
    for producing a netlist in an in-memory string.
    Format-specifier `fmt` may be any of the `NetlistFormatSpec` enumerated values
    or their string equivalents.
    """

    if opts is not None:
        raise NotImplementedError("NetlistOptions not yet implemented.")  # FIXME!

    # If `fmt` is a string, turn it into an enum
    fmt_enum = NetlistFormat.get(fmt)

    # Get the corresponding `Netlister` class and instantiate it
    netlister_cls = fmt_enum.netlister()
    netlister = netlister_cls(dest=dest)

    # Write the netlist
    return netlister.write_package(pkg)


# Set our exported content for star-imports
__all__ = ["netlist", "NetlistFormat", "NetlistFormatSpec"]



"""
Spectre Implementation of `vlsir.spice.Sim`
"""

# Std-Lib Imports
import subprocess, re, shutil, glob
import numpy as np
from typing import Tuple, Any, Mapping, Optional, IO, Dict
from dataclasses import dataclass
from warnings import warn
from enum import Enum

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from ..netlist.spectre import SpectreNetlister
from .base import Sim
from .sim_data import TranResult, OpResult, SimResult, AcResult, DcResult
from .spice import SupportedSimulators, sim

# Module-level configuration. Over-writeable by sufficiently motivated users.

# The simulator executable invoked. If over-ridden, likely for sake of a specific path or version.
SPECTRE_EXECUTABLE = "spectre"
# Additional arguments to pass to the simulator executable.
SPECTRE_ARGS = ""  ##"++aps"


def available() -> bool:
    return SpectreSim.available()


class SpectreSim(Sim):
    """
    State and execution logic for a Spectre-call to `vsp.Sim`.
    """

    @staticmethod
    def available() -> bool:
        """Boolean indication of whether the current running environment includes the simulator executable."""
        if shutil.which(SPECTRE_EXECUTABLE) is None:
            return False
        try:
            # And if it's set, check that we can get its version without croaking.
            # This can often happen because of an inaccessible license server, or just a badly-linked installation.
            subprocess.run(
                f"{SPECTRE_EXECUTABLE} -V",  # Yes, "version" gets a capital "V" for this program (ascii shrug)
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=10,
            )
        except Exception:
            # Indicate "not available" for any Exception. Usually this will be a `subprocess.CalledProcessError`.
            return False
        return True  # Otherwise, installation looks good.

    @classmethod
    def enum(cls) -> SupportedSimulators:
        return SupportedSimulators.SPECTRE

    def run(self) -> SimResult:
        """Run the specified `SimInput` in directory `self.rundir`, returning its results."""

        # Write our netlist to file
        self.write_netlist()

        # Run the simulation
        self.run_spectre_process()

        # Parse the results
        return self.parse_results()

    def write_netlist(self) -> None:
        """# Write our netlist to file"""

        netlist_file = self.open("netlist.scs", "w")
        netlister = SpectreNetlister(dest=netlist_file)
        netlister.write_sim_input(self.inp)
        netlist_file.close()

    def parse_results(self) -> SimResult:
        """# Parse output data"""

        # Parse output data
        data = parse_nutbin(self.open("netlist.raw", "rb"))
        an_type_dispatch = dict(
            ac=self.parse_ac, dc=self.parse_dc, op=self.parse_op, tran=self.parse_tran
        )
        results = []
        for an in self.inp.an:
            an_type = an.WhichOneof("an")
            inner = getattr(an, an_type)
            if an_type not in an_type_dispatch:
                msg = f"Invalid or Unsupported analysis {an} with type {an_type}"
                raise RuntimeError(msg)
            func = an_type_dispatch[an_type]
            if inner.analysis_name not in data:
                msg = f"Cannot read results for analysis {an}"
                raise RuntimeError(msg)
            inner_data = data[inner.analysis_name]
            an_results = func(inner, inner_data)
            results.append(an_results)

        return SimResult(an=results)

    def parse_ac(self, an: vsp.AcInput, nutbin: "NutBinAnalysis") -> AcResult:
        # FIXME: the `mt0` and friends file names collide with tran, if they are used in the same Sim!
        measurements = self.get_measurements("*.mt*")

        # Pop the frequence vector out of the data
        freq = nutbin.data.pop("freq")
        # Nutbin format stores the frequency vector as complex numbers, along with all the complex-valued signal data.
        # Grab the real parts of the frequencies, and ensure that they don't (somehow) have nonzero imaginary parts.
        if np.any(freq.imag):
            raise RuntimeError(f"Imaginary parts of frequencies in {freq}")
        freq = freq.real

        return AcResult(
            analysis_name=an.analysis_name,
            freq=freq,
            data=nutbin.data,
            measurements=measurements,
        )

    def parse_dc(self, an: vsp.DcInput, nutbin: "NutBinAnalysis") -> DcResult:
        measurements = self.get_measurements("*.ms*")
        return DcResult(
            analysis_name=an.analysis_name,
            indep_name=an.indep_name,
            data=nutbin.data,
            measurements=measurements,
        )

    def parse_op(self, an: vsp.OpInput, nutbin: "NutBinAnalysis") -> OpResult:
        return OpResult(
            analysis_name=an.analysis_name,
            data={k: v[0] for k, v in nutbin.data.items()},
        )

    def parse_tran(self, an: vsp.TranInput, nutbin: "NutBinAnalysis") -> TranResult:
        """Extract the results for Analysis `an` from `data`."""
        measurements = self.get_measurements("*.mt*")
        return TranResult(
            analysis_name=an.analysis_name, data=nutbin.data, measurements=measurements
        )

    def get_measurements(self, filepat: str) -> Dict[str, float]:
        """Get the measurements at files matching (glob) `filepat`.
        Returns only a single files-worth of measurements, and issues a warning if more than one such file exists.
        Returns an empty dictionary if no matching files are found."""
        meas_files = list(self.glob(filepat))
        if not meas_files:
            return dict()
        if len(meas_files) > 1:
            msg = f"Unsupported: more than one measurement-file generated. Only the first will be read"
            warn(msg)
        return parse_mt0(self.open(meas_files[0], "r"))

    def run_spectre_process(self) -> None:
        """Run a Spectre sub-process, executing the simulation"""
        # Note the `nutbin` output format is dictated here
        cmd = (
            f"{SPECTRE_EXECUTABLE} {SPECTRE_ARGS} -E -format nutbin netlist.scs".split(
                " "
            )
        )
        return self.run_subprocess(cmd)


class FromStr:
    """Mix-in for creating `Enum`s from string values."""

    @classmethod
    def from_str(cls, s: str) -> "FromStr":
        reversed = {v.value: v for v in cls}
        if s not in reversed:
            msg = f"Invalid `{cls.__name__}`: `{s}`"
            raise ValueError(msg)
        return reversed[s]


class NumType(FromStr, Enum):
    """# Numeric Datatypes
    Values equal the strings used in NutBin data files."""

    REAL = "real"
    COMPLEX = "complex"


class Units(FromStr, Enum):
    """# Unit Types
    Values equal the strings used in NutBin data files."""

    DUMMY = "dummy"
    SWEEP = "sweep"
    SECONDS = "s"
    VOLTS = "V"
    AMPS = "A"
    HERTZ = "Hz"
    CELSIUS = "C"


@dataclass
class VarSpec:
    """Variable Spec"""

    name: str
    units: Units


@dataclass
class NutBinAnalysis:
    """Analysis result from a NutBin file."""

    analysis_name: str  # Analysis name
    numtype: NumType  # Numeric type in `data` field
    data: Mapping[str, np.ndarray]  # Signal name => data
    units: Mapping[str, Units]  # Signal name => units


def parse_nutbin(f: IO) -> Mapping[str, NutBinAnalysis]:
    """Parse a `nutbin` format set of simulation results.
    Returns results as a dictionary from `analysis_name` to `NutBinAnalysis`.
    Note this is paired with the simulator invocation commands, which include `format=nutbin`."""

    # Parse per-file header info
    # First 2 lines are ascii one line statements
    _title = f.readline()  # Title, ignored
    _date = f.readline()  # Run date, ignored

    # And parse the rest of the file per-analysis
    rv = {}
    while f:
        plotname = f.readline().decode("ascii")  # Simulation name
        if len(plotname) == 0:
            break
        an = parse_nutbin_analysis(f, plotname)
        rv[an.analysis_name] = an
    return rv


def parse_nutbin_analysis(f: IO, plotname: str) -> NutBinAnalysis:
    """Parse a `NutBinAnalysis` from an open nutbin-format file `f`."""

    # Next 4 lines are also ascii one line statements
    # Parse the `Flags` field, which primarily includes the numeric datatype
    flags = f.readline().decode("ascii")
    flags = flags.split()
    if len(flags) != 2 or flags[0] != "Flags:":
        raise ValueError(f"Invalid flags {flags}")
    numtype = NumType.from_str(flags[1])
    nptypes = {NumType.REAL: float, NumType.COMPLEX: complex}
    nptype = nptypes[numtype]

    # Parse the number of variables and points
    num_vars_line = f.readline()  # No. Variables:   [nvar]
    num_pts_line = f.readline()  # No. Points:      [npts]
    sim_name = plotname.split("`")[-1].split("'")[0]

    # Find the number of variables and number of points
    num_vars = int(
        re.match(
            r"No. Variables:\s+(?P<num_vars>\d+)\n",
            num_vars_line.decode("ascii"),
        ).group("num_vars")
    )
    num_pts = int(
        re.match(
            r"No. Points:\s+(?P<num_pts>\d+)\n",
            num_pts_line.decode("ascii"),
        ).group("num_pts")
    )

    # Decode the variables spec, looks like the following
    # Variables: [Variable idx] [Variable name] [units] [optional_flags]
    var_line = f.readline().decode("ascii")
    var_specs = [_read_var_spec(var_line[10:])]
    for i in range(num_vars - 1):
        var_specs.append(_read_var_spec(f.readline().decode("ascii")))

    # Read the binary data, should look like the following:
    # Binary: \n[Binary data]
    binary_line = f.readline().decode("ascii")
    assert binary_line == "Binary:\n"
    # Data is big endian
    bin_data = np.fromfile(
        f, dtype=np.dtype(nptype).newbyteorder(">"), count=num_vars * num_pts
    )
    data = {}
    units = {}
    for i, var in enumerate(var_specs):
        data[var.name] = bin_data[i::num_vars]
        units[var.name] = var.units

    return NutBinAnalysis(
        analysis_name=sim_name,
        numtype=numtype,
        data=data,
        units=units,
    )


def _read_var_spec(line: str) -> VarSpec:
    """Read a Variable spec line from the input
    and return the name and the units."""
    m = re.match(
        r"\s+(?P<idx>\d+)\s+(?P<name>\S+)\s+(?P<units>\S+)(?P<rest>.*)\n", line
    )
    return VarSpec(name=m.group("name"), units=Units.from_str(m.group("units")))


def parse_mt0(file: IO) -> Dict[str, float]:
    """Parse an (open) "mt0-format" measurement-file into a set of {name: value} pairs."""

    file.readline()  # Header
    file.readline()  # Netlist Title
    keys = file.readline()  # Measurement Names Line
    keys = keys.split()
    values = file.readline()  # Measurement Values Line

    def convert(s: str) -> float:
        """Convert a string to a float, converting failing cases to `NaN`"""
        try:
            return float(s)
        except:
            return float("NaN")

    values = [convert(s) for s in values.split()]
    return dict(zip(keys, values))



"""
# SPICE Types

The "spice type system", in the form of prefix-characters for instances. 
Implemented as a Python-native `Enum`, 
in which each value is the prefix character. 
Includes methods for conversion to and from the VLSIR schema `enum`. 
"""

# Std-Lib Imports
from enum import Enum

# VLSIR Imports
from vlsir.circuit_pb2 import SpiceType as SchemaSpiceType


class SpiceType(Enum):
    """# Spice Type
    Enumerated Spice Types and their Instance-Name Prefixes"""

    # Sub-circits, either from `Module`s or `ExternalModule`s
    SUBCKT = "x"
    # Ideal Passives
    RESISTOR = "r"
    CAPACITOR = "c"
    INDUCTOR = "l"
    # Semiconductor Devices
    MOS = "m"
    DIODE = "d"
    BIPOLAR = "q"
    # Independent Sources
    VSOURCE = "v"
    ISOURCE = "i"
    # Dependent Sources
    VCVS = "e"
    VCCS = "g"
    CCCS = "f"
    CCVS = "h"
    # Transmission Lines
    TLINE = "o"

    def to_schema(self) -> SchemaSpiceType:
        """Convert to VLSIR protobuf schema"""

        if self == SpiceType.SUBCKT:
            return SchemaSpiceType.SUBCKT
        if self == SpiceType.RESISTOR:
            return SchemaSpiceType.RESISTOR
        if self == SpiceType.CAPACITOR:
            return SchemaSpiceType.CAPACITOR
        if self == SpiceType.INDUCTOR:
            return SchemaSpiceType.INDUCTOR
        if self == SpiceType.MOS:
            return SchemaSpiceType.MOS
        if self == SpiceType.DIODE:
            return SchemaSpiceType.DIODE
        if self == SpiceType.BIPOLAR:
            return SchemaSpiceType.BIPOLAR
        if self == SpiceType.VSOURCE:
            return SchemaSpiceType.VSOURCE
        if self == SpiceType.ISOURCE:
            return SchemaSpiceType.ISOURCE
        if self == SpiceType.VCVS:
            return SchemaSpiceType.VCVS
        if self == SpiceType.VCCS:
            return SchemaSpiceType.VCCS
        if self == SpiceType.CCCS:
            return SchemaSpiceType.CCCS
        if self == SpiceType.CCVS:
            return SchemaSpiceType.CCVS
        if self == SpiceType.TLINE:
            return SchemaSpiceType.TLINE
        raise ValueError(f"Invalid SpiceType: {self}")

    @staticmethod
    def from_schema(schema: SchemaSpiceType) -> "SpiceType":
        """Create from VLSIR schema"""

        if schema == SchemaSpiceType.SUBCKT:
            return SpiceType.SUBCKT
        if schema == SchemaSpiceType.RESISTOR:
            return SpiceType.RESISTOR
        if schema == SchemaSpiceType.CAPACITOR:
            return SpiceType.CAPACITOR
        if schema == SchemaSpiceType.INDUCTOR:
            return SpiceType.INDUCTOR
        if schema == SchemaSpiceType.MOS:
            return SpiceType.MOS
        if schema == SchemaSpiceType.DIODE:
            return SpiceType.DIODE
        if schema == SchemaSpiceType.BIPOLAR:
            return SpiceType.BIPOLAR
        if schema == SchemaSpiceType.VSOURCE:
            return SpiceType.VSOURCE
        if schema == SchemaSpiceType.ISOURCE:
            return SpiceType.ISOURCE
        if schema == SchemaSpiceType.VCVS:
            return SpiceType.VCVS
        if schema == SchemaSpiceType.VCCS:
            return SpiceType.VCCS
        if schema == SchemaSpiceType.CCCS:
            return SpiceType.CCCS
        if schema == SchemaSpiceType.CCVS:
            return SpiceType.CCVS
        if schema == SchemaSpiceType.TLINE:
            return SpiceType.TLINE
        raise ValueError(f"Invalid SchemaSpiceType: {schema}")



"""
# Setup Script

Derived from the setuptools sample project at
https://github.com/pypa/sampleproject/blob/main/setup.py

"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

# Get the long description from the README file
here = pathlib.Path(__file__).parent.resolve()
readme = here / "readme.md"
long_description = "" if not readme.exists() else readme.read_text(encoding="utf-8")

_VLSIR_VERSION = "4.0.0"

setup(
    name="vlsirtools",
    version=_VLSIR_VERSION,
    description="Tools for the Vlsir IC Design Schema",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Vlsir/Vlsir",
    author="Dan Fritchman",
    author_email="dan@fritch.mn",
    packages=find_packages(),
    python_requires=">=3.7, <3.12",
    install_requires=[
        f"vlsir=={_VLSIR_VERSION}",  # VLSIR Core Python Bindings
        "numpy~=1.21",  # For `sim_data` simulation results
        "pandas~=1.3",  # For CSV reading
    ],
    extras_require={
        "dev": ["pytest==7.1", "coverage", "pytest-cov", "black==22.6", "twine"]
    },
)
